#!/usr/bin/python

__author__      = "Sander Granneman"
__copyright__   = "Copyright 2018"
__version__     = "0.0.4"
__credits__     = ["Sander Granneman"]
__maintainer__  = "Sander Granneman"
__email__       = "sgrannem@staffmail.ed.ac.uk"
__status__      = "Production"

##################################################################################
#
#   pyCRAC NGSFormatWriters.py
#
#
#   Copyright (c) Sander Granneman 2019
#   
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#   
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#   
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#
##################################################################################

import sys
from pyCRAC.Classes.Exceptions import *

class NGSFileWriter():
    """ Converts transposed numpy arrays in to various NGS output formats """
    def __init__(self,outfile=None):
        self.attributestring = str()
        if not outfile:
            self.outfile = sys.stdout
        elif hasattr(outfile,'read'):
            self.outfile = outfile
        else:
            self.outfile = open(outfile,"w")
                    
    def rgbColorCodes(self,color):
        """ returns the RGB code for the color of interest. More or less randomly picked colors from a website """
        colorcode = {
                "white" : "255,255,255",
                "black" : "0,0,0",
                "gray"  : "190,190,190",
                "blue"  : "0,0,255",
                "cyan"  : "0,255,255",
                "aquamarine" : "127,255,170",
                "green" : "0,100,0",
                "yellow": "255,255,0",
                "gold"  : "218,165,32",
                "brown" : "139,69,19",
                "beige" : "245,245,220",
                "wheat" : "245,222,179",
                "chocolate" : "210,105,30",
                "salmon": "233,150,122",
                "orange": "255,165,0",
                "tomato": "255,99,71",
                "red"   : "255,0,0",
                "pink"  : "255,20,147",
                "plum"  : "221,160,221",
                "violet": "238,130,238",
                "purple": "160,32,240"
        }
        if color in colorcode:
            return colorcode[color]
        else:
            raise LookupError("\ncould not find the color in the colorcode dictionary.\nPlease choose from the following:\n%s\n" % "\n".join(colorcode.keys()))

    def write(self,string):
        """ writes any string to the output file """
        self.outfile.write(string)
    
    def close(self):
        """ closes the output file """
        self.outfile.close()

    def writeTrackLine(self,track_type="bed",name="User_supplied",description="User_supplied",colorbystrand="red,blue",color="black"):
        """ writes a track line for the UCSC genome browser. Note that UCSC has poor support for the sgr format. The colorbystrand 
        variable indicates the colors for each strand. "red,blue" means that features located on the '+' strand will be colored
        red and features on the '-' strand will be colored blue.
        Example: track type=bed name='User_supplied_track' description='User_supplied_track' colorByStrand='225,0,0 0,0,225'"""
        types = ["bed","bedGraph","gtf","gff"]
        color = self.rgbColorCodes(color)
        twocolors     = colorbystrand.split(",")
        colorbystrand = "%s %s" % (self.rgbColorCodes(twocolors[0]),self.rgbColorCodes(twocolors[1]))
        if track_type not in types:
            raise LookupError("\ncould not find %s in the types array.\nPlease choose from the following:\n%s\n" % (track_type,", ".join(types)))
        self.outfile.write("track type=%s name='%s' description='%s' color='%s' colorByStrand='%s'\n" % (track_type,name,description,color,colorbystrand))             
                    
    def writeBedgraph(self,chromosome,start,end,hits):
        """ writes a bedgraph output string to a file. Note that these are 1-based coordinates."""
        self.outfile.write("%s\t%s\t%s\t%s\n" % (chromosome,start+1,end+1,hits))
            
    def writeBed(self,chromosome,start,end,strand,name=".",score="."):
        """ writes a bed output string to a file. Note that these are 0-based coordinates."""
        self.outfile.write("%s\t%s\t%s\t%s\t%s\t%s\n" % (chromosome,start,end,name,score,strand))

    def writeSgr(self,chromosome,position,hits):
        """ takes data in transposed numpy arrays and writes it in sgr format to a specified output file. Only allows data from one chromosome and strand.
        Note that these are 1-based coordinates."""
        self.outfile.write("%s\t%s\t%s\n" % (chromosome,position+1,hits)) 

    def writeGTF(self,seqname,source,feature,start,end,score=".",strand=".",frame=".",gene_name=None,transcript_name=None,gene_id=None,transcript_id=None,exon_number=None,comments=None):  
        """method to feed the individual features of the GTF file to the writer, which prints the variables into a string. Note that these are 1-based coordinates."""
        self.attributestring = str()
        #for i in [start,end]:
        #    if not isinstance(i,int) or not isinstance(i,float):
        #    	sys.stderr.write("seqname:\t%s\nsource:\t%s\nstart:\t%s\nend:\t%s\nscore:\t%s\n" % (seqname,source,start,end,score))
        #        raise TypeError("\n%s position needs to be an integer or a float\n" % i)
        if gene_id:         self.attributestring += "gene_id \"%s\"; " % gene_id
        if gene_name:       self.attributestring += "gene_name \"%s\"; " % gene_name
        if transcript_id:   self.attributestring += "transcript_id \"%s\"; " % transcript_id
        if transcript_name: self.attributestring += "transcript_name \"%s\"; " % transcript_name
        if exon_number:     self.attributestring += "\"%s\"; " % exon_number
        if comments:        self.attributestring += "# %s;" % comments
        if not self.attributestring: self.attributestring = "."
        self.outfile.write("%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % (seqname,source,feature,start+1,end,score,strand,frame,self.attributestring))

    def writeGFF(self,seqname,source,feature,start,end,score=".",strand=".",frame=".",gene_name=None,transcript_name=None,gene_id=None,transcript_id=None,exon_number=None):  
        """method to feed the individual features of the GTF file to the writer, which prints the variables into a string. Note that these are 1-based coordinates."""
        self.attributestring = str()
        #for i in [start,end]:
        #    if not isinstance(i,int) or not isinstance(i,float):
        #    	sys.stderr.write("seqname:\t%s\nsource:\t%s\nstart:\t%s\nend:\t%s\nscore:\t%s\n" % (seqname,source,start,end,score))
        #        raise TypeError("\n%s position needs to be an integer or a float\n" % i)
        if gene_id:         self.attributestring += "ID=\"%s\"; " % gene_id
        if gene_name:       self.attributestring += "name=\"%s\"; " % gene_name
        if transcript_id:   self.attributestring += "Alias=\"%s\"; " % transcript_id
        if not self.attributestring: self.attributestring = "."
        self.outfile.write("%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % (seqname,source,feature,start+1,end,score,strand,frame,self.attributestring))



#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "0.0.4"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyFilterGTF.py
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
from optparse import *
from pyCRAC.Parsers import GTF2
from pyCRAC.Classes.NGSFormatWriters import *
from pyCRAC.Parsers.ParseAlignments import ParseCountersOutput

def filterGenes(genes,gtf,strand,orientation="sense",singlefeats=False):
	""" filters genes on whether they are sense or anti-sense to a feature """
	filteredgenes = list()
	for gene in genes:
		if gene in gtf.genes:
			if orientation == "sense":
				if gtf.strand(gene) == strand:
					filteredgenes.append(gene)
			else:
				if gtf.strand(gene) != strand:
					filteredgenes.append(gene)
			if singlefeats and len(filteredgenes) > 1:
				filteredgenes = list()
	return filteredgenes

def main():
	parser = OptionParser(usage="usage: %prog [options] -f filename", version="%s" % __version__)
	files = OptionGroup(parser, "File input options")
	files.add_option("-f", "--input_file", dest="filename",help="provide the path to your gtf data file. Default is standard input.", metavar="FILE",default=None)
	files.add_option("-o", "--output_file", dest="output_file",help=" Use this flag to provide an output file name. Default is standard output. ", default=None)
	files.add_option("--gtf",dest="annotation_file",metavar="annotation_file.gtf",help="type the path to the gtf annotation file that you want to use",default=None)
	filtering = OptionGroup(parser, "GTF filtering options")
	filtering.add_option("--singlefeats",dest="singlefeats",action="store_true",help="to remove intervals that were mapped to multiple gene names, sense or anti-sense. Default = False",default=False)
	filtering.add_option("-a","--annotation",dest="annotation",help="select which annotation (i.e. protein_coding, ncRNA, sRNA, rRNA, tRNA, snoRNA) you would like to focus your search on. Default = None.", metavar="protein_coding",default=None)
	filtering.add_option("--anti_sense",dest="anti_sense",action="store_true",help="by default pyFilterGTF.py only reports intervals that overlap 'sense' with features. Use this flag if you only want to plot intervals that map anti-sense.",default=False)
	parser.add_option_group(files)
	parser.add_option_group(filtering)
	(options, args) = parser.parse_args()
	
	### setting the orientation
	orientation = "sense"
	if options.anti_sense:
		orientation = "anti_sense"
	###

	### By default, input and output are expected from the standard input or standard output.
	infile = sys.stdin
	outfile = sys.stdout
	###
	
	### Making GTF2 parser object and parsing GTF annotation file ###
	gtf = GTF2.Parse_GTF()
	gtf.read_GTF(options.annotation_file,source=options.annotation,transcripts=False)
	###
	
	### Object for processing the GTF data files
	if options.filename:
		infile = open(options.filename,"r")
	datain = ParseCountersOutput(infile)
	###
	
	### Object for writing the GTF output filename
	if options.output_file:
		outfile = open(options.output_file,"w")
	dataout = NGSFileWriter(outfile)
	###
	
	while datain.readLineByLine(numpy=False,mutsseqpos=False,collectmutations=False):
		genes = filterGenes(datain.genes,gtf,datain.strand,orientation=orientation,singlefeats=options.singlefeats)
		if genes:
			dataout.writeGTF(datain.chromosome,datain.source,datain.type,datain.read_start,datain.read_end,datain.score,datain.strand,gene_name=",".join(genes),gene_id=",".join([gtf.gene2orf(i) for i in genes]),comments=datain.comments)		
	
if __name__ == "__main__":
	main()


#!/usr/bin/python

__authors__		= ["Sander Granneman"]
__copyright__	= "Copyright 2019"
__version__		= "3.1"
__credits__		= ["Sander Granneman","Grzegorz Kudla"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyBarcodeFilter.py
#
#
#	Copyright (c) Sander Granneman and Grzegorz Kudla 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import re
import time
from optparse import *
from pyCRAC.Classes.Barcodes import *
					
def main():
	parser = OptionParser(usage="usage: %prog [options] -f filename -b barcode_list", version="%s" % __version__)
	parser.add_option("-f", "--input_file", dest="filename",help="name of the FASTQ or FASTA input file", metavar="FILE")
	parser.add_option("-r", "--reverse_input_file", dest="filename_reverse",help="name of the paired (or reverse) FASTQ or FASTA input file", metavar="FILE")
	parser.add_option("--file_type", dest="filetype",help="type of file, uncompressed (fasta or fastq) or compressed (fasta.gz or fastq.gz, gzip/gunzip compressed). Default is fastq", metavar="FASTQ",default="fastq")
	parser.add_option("-b", "--barcode_list", dest="barcode_list",help="name of tab-delimited file containing barcodes and barcode names", metavar="FILE",default=None)
	parser.add_option("-k", "--keep_barcode", dest= "keep_barcode", action="store_true", help="in case you do not want to remove the in read barcode from the sequences. Useful when uploading data to repositories.",default=False)
	parser.add_option("-v","--verbose",dest="verbose",action="store_true",help="prints all the status messages to a file rather than the standard output.",default=False)
	parser.add_option("--search",choices=["forward","reverse"],help="use this flag if barcode sequences need to be removed from the forward or the reverse reads. The tool assumes. Default=forward", default="forward")
	parser.add_option("-m", "--mismatches", dest="mismatches",help="to set the number of allowed mismatches in a barcode. Only one mismatch is allowed. Default = 0", type="int",metavar="1",default=0)
	parser.add_option("-i", "--index", dest="index",help="use this option if you want to split the data using the Illumina indexing barcode information",action="store_true",default=False)
	parser.add_option("--gz","--gzip", dest="gzip", action="store_true",help="use this option to compress all the output files using gunzip or gzip (compression level 9). Note that this can significantly slow down the program", default=False)
	(options, args) = parser.parse_args()
	data = None
	if len(sys.argv) < 4:
		parser.error("At least two file arguments are required. Use -h option for usage\n")
	if not options.barcode_list:
		parser.error("You forgot to enter the name of the file containing the barcodes. Use the -b flag for this purpose\n")
	if not options.filename:
		parser.error("You forgot to enter the name of the (forward) raw data file. Use the -f flag for this purpose\n")
	if options.index:
		data = IndexingBarcodeSplitter(options.filetype,options.filename,options.filename_reverse,options.barcode_list,allowedmismatches=options.mismatches,gzip=options.gzip)
		if options.filename_reverse:
			if re.search("fastq",options.filetype):
				data.processPairedFastqFileIndexes()
			elif re.search("fasta",options.filetype):
				data.processPairedEndFastAFileIndexes()
			else:
				parser.error("\nDid not recognize the %s file type\n" % options.filetype)
		else:
			if re.search("fastq",options.filetype):
				data.processFastqFileIndexes()
			elif re.search("fasta",options.filetype):
				data.processFastAFileIndexes()
			else:
				parser.error("\nDid not recognize the %s file type\n" % options.filetype)

	else:
		data = BarcodeSplitter(options.filetype,options.filename,options.filename_reverse,options.barcode_list,allowedmismatches=options.mismatches,gzip=options.gzip,keepbarcode=options.keep_barcode,search=options.search)
		if options.filename_reverse:
			if re.search("fastq",options.filetype):
				data.processPairedEndFastQFiles()
			elif re.search("fasta",options.filetype):
				data.processPairedEndFastAFiles()
			else:
				parser.error("\nDid not recognize the %s file type\n" % options.filetype)
		else:
			if re.search("fastq",options.filetype):
				data.processFastQFile()
			elif re.search("fasta",options.filetype):
				data.processFastAFile()
			else:
				parser.error("\nDid not recognize the %s file type\n" % options.filetype)

	data.printRandBarcodeStats()
	data.printBarcodeStats()

if __name__ == "__main__":
	main()


#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2020"
__version__		= "0.0.9"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@staffmail.ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	FDR.py
#
#
#	Copyright (c) Sander Granneman 2020
#
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import numpy as np
import tempfile
from pyCRAC.Methods import numpy_overlap
from pyCRAC.Classes.NGSFormatReaders import NGSFileReader
from collections import defaultdict

class ModFDR():
	""" class for calculating False Discovery Rates (FDRs) for NGS data intervals. Inspired by Pyicos
	(http://regulatorygenomics.upf.edu/group/media/pyicos_docs/index.html) but implemented using numpy
	allowing fast numerical calculations on large arrays """
	def __init__(self,iterations=100,mincoverage=1,debug=False,minfdr=0.05):
		limit = 1000
		assert iterations <=limit, "To preserve memory the iteration maximum is set to %s\n" % limit
		self.iterations	  = iterations
		self.mincoverage  = mincoverage
		self.debug		  = debug
		self.minfdr		  = minfdr
		self.FDR		  = float()
		self.FDR_counter  = defaultdict(lambda: defaultdict(list))
		self.coverage	  = int()

	def writeFDRData(self,log):
		""" writes coverage and FDR information to a log file. Useful for debugging purposes.
		Resets the FDR_counter dict when finished. """
		assert self.FDR_counter, "No data to be reported\n"
		for fdr in sorted(self.FDR_counter["FDR"]):
			if len(self.FDR_counter["FDR"][fdr]) > 1:
				coverage = min(self.FDR_counter["FDR"][fdr])
				actual	 = sum([self.FDR_counter["actual_coverage"][i] for i in self.FDR_counter["FDR"][fdr]])
				control	 = sum([self.FDR_counter["control_coverage"][i] for i in self.FDR_counter["FDR"][fdr]])
				log.write(">=%s\t%s\t%s\t%s\n" % (coverage,fdr,actual,control))
			else:
				coverage = self.FDR_counter["FDR"][fdr][0]
				actual	 = self.FDR_counter["actual_coverage"][coverage]
				control	 = self.FDR_counter["control_coverage"][coverage]
				log.write("%s\t%s\t%s\t%s\n" % (coverage,fdr,actual,control))

	def findSignificantRegions(self,actual,control,randomised=True):
		""" calculates False discovery rates (FDRs) for a region (i.e. gene positions) by comparing the experimental data to randomised data.
		It then masks regions within the actual data that fall below the set minimal FDR.
		If you provide an array with negative control data, then you have to set the 'randomised' variable to 'False'.
		Requires two numpy arrays with experimental data and one numpy array with randomized or control data. The randomised
		data should be a matrix in which each row represents the randomized coverage for each iteration.
		Returns the results as a dictionary in which each keys are heights and the values are FDRs """
		self.FDR_counter = defaultdict(lambda: defaultdict(list))	# the counter is reset every time to preserve memory
		region_length = float(len(actual))
		max_coverage  = int(actual.max())
		if max_coverage >= self.mincoverage:
			for height in reversed(list(range(1,max_coverage+1))):
				p_actual = (actual >= height).sum(dtype=float)/region_length			# what is the probability of coverage 'height' in the actual data?
				self.FDR_counter["actual_coverage"][height] = (actual >= height).sum()	# stores occurrences of height in actual data
				if randomised:															# if the data is randomised read positions, calculate the probabilities and standard deviations for the randomised data
					p_rand = (control >= height).sum(axis=1,dtype=float)/region_length	# what is the probability of coverage 'height' in each of the random iterations?
					if not p_rand.any():												# if p_rand is zero then there is no point in calculating FDRs...
						self.FDR = 0.0
					else:
						p_rand_mean = p_rand.mean(dtype=float)							# mean of the random probabilities for coverage 'height'
						p_rand_std	= p_rand.std(dtype=float)							# standard deviation of the random probabilities for coverage 'height'
						self.FDR	= p_rand_mean + p_rand_std / p_actual
					self.FDR_counter["control_coverage"][height] = (control >= height).sum()/float(self.iterations) # stores occurrences of height in randomized data

				else:																	# if the data is generated from a negative control dataset, calculate the probability using the control data
					p_control = (control >= height).sum(dtype=float)/region_length		# what is the probability of coverage 'height' in the control data?
					if not p_control.any(): 											# if p_control is zero then there is no point in calculating FDRs...
						self.FDR = 0.0
					else:
						self.FDR  = p_control / p_actual
					self.FDR_counter["control_coverage"][height] = (control >= height).sum()	# stores occurrences of height in control data
				self.FDR_counter["FDR"][self.FDR].append(height)						# stores FDR value for each height
				if self.FDR >= self.minfdr and height + 1 >= self.mincoverage:
					self.coverage = height+1
					return np.where(actual >= self.coverage)[0]

		return np.zeros(1)	# If the coverage is below the mincoverage threshold or it could not find significant regions it returns an empty numpy list

	def calculatePeakHeightFDR(self,actual,control,peakheight):
		""" For a specific peak height it will calcualte the FDR """
		region_length = float(len(actual))
		p_actual  = (actual >= peakheight).sum(dtype=float)/region_length			# what is the probability of coverage 'height' in the actual data?
		p_control = (control >= peakheight).sum(axis=1,dtype=float)/region_length		# what is the probability of coverage 'height' in each of the random iterations?
		p_control_mean = p_control.mean(dtype=float)									# mean of the random probabilities for coverage 'height'
		p_control_std  = p_control.std(dtype=float)									# standard deviation of the random probabilities for coverage 'height'
		fdr	= p_control_mean + p_control_std / p_actual
		return fdr

	def shuffleReadPositions(self,regionlength,regionreadinfo):
		""" generates a matrix with shape genelength by number of iterations and randomizes read positions (regionreadinfo dict) over each row.
		Returns the matrix with randomized data. The regionreadinfo dictionary should have information about the number of reads that have a particular read length.
		for example regionreadinfo[10] = 5 indicates that the region has 5 reads of read length 10. This information is required for shuffling the read positions over the region of interest. """
		randarray = np.zeros(shape=(self.iterations,regionlength))						# make a matrix with a number of rows, dictated by number_of_iterations
		for readlength in regionreadinfo:												# genereadinfo[readlengths] = 5
			for i in range(regionreadinfo[readlength]):
				random_numbers = np.random.randint(-readlength+1,regionlength,self.iterations)	# generate random numbers for each row as starting positions. I wonder if a better distribution (other than the normal distribution) can be used here
				for j in range(self.iterations):
					if random_numbers[j] < 0:
						randarray[j][:random_numbers[j]+readlength] += 1
					else:
						randarray[j][random_numbers[j]:random_numbers[j]+readlength] += 1

		return randarray

	def processReadData(self,gtf,datafile,chromosomelength,type="gtf",todisk=False,ignorestrand=False,tempdirectory="./"):
		""" processes read data from a single chromosome and calculates for each position overlapping sense with a genomic feature using numpy arrays.
		Requires a GTF2.Parse_GTF instance loaded with GTF annotations. Supports bed6, gtf and gff as input files."""
		typechoices = ["bed","gtf","gff"]
		data = NGSFileReader(datafile)
		functions = [data.readBedLine,data.readGTFLine,data.readGFFLine]
		reader = functions[typechoices.index(type)]
		chromarray = defaultdict(list)
		chromarray["+"] = np.zeros(chromosomelength)
		chromarray["-"] = np.zeros(chromosomelength)
		list_of_tuples = defaultdict(list)
		genecoverageinfo = defaultdict(lambda: defaultdict(int))
		score = 1
		while reader():
			if (data.chromosome,data.strand) not in list_of_tuples:
				try:
					list_of_tuples[data.chromosome,data.strand] = gtf.chromosomeGeneCoordIterator(data.chromosome,numpy=True,strand=data.strand)	# I am only allowing 'sense' overlap
				except AssertionError:
					continue
			search_results = numpy_overlap(list_of_tuples[data.chromosome,data.strand],data.start,data.end-1,overlap=1)
			if search_results:																		# only add to the chromarray if overlap was found with a genomic feature
				if data.strand == "." or ignorestrand:												# if the feature has no strand or the data has no strand, put everything in the plus strand array.
					data.strand = "+"
				readlength = data.end - data.start
				if data.score.isdigit(): score = int(data.score)
				chromarray[data.strand][data.start:data.end] += score
				for gene in search_results:
					genecoverageinfo[gene][readlength] += score
		if todisk:																					# not fully tested and experimental
			tempplus  = tempfile.NamedTemporaryFile(mode='w',suffix=".npy",dir=tempdirectory,delete=False)	# making temp files
			tempminus = tempfile.NamedTemporaryFile(mode='w',suffix=".npy",dir=tempdirectory,delete=False)
			np.save(tempplus,chromarray["+"])
			np.save(tempminus,chromarray["-"])
			chromarray["+"] = np.memmap(tempplus.name,shape=(chromosomelength))						# each chromarray key is linked to a numpy.memmap
			chromarray["-"] = np.memmap(tempminus.name,shape=(chromosomelength))
			chromarray["files"] = list()
			chromarray["files"].extend([tempplus,tempminus])
		return chromarray,genecoverageinfo



#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "0.1.1"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyGTF2BedGraph.py
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import numpy as np
from collections import defaultdict
from optparse import *
from pyCRAC.Parsers.ParseAlignments import ParseCountersOutput
from pyCRAC.Methods import getfilename,processChromFile,steparray2Intervals
from pyCRAC.Classes.NGSFormatWriters import *

def gtf2bedgraph(datafile,chromosomedata,out_files=[],type="reads",permillion=False,fiveend=False,threeend=False,score=False,log=sys.stdout,name=None,description=None,colorbystrand="red,blue"):
	"""Produces a bedgraph output file for hits, substitutions and deletions. It requires, besides data a file containing 
	chromosome names and chromosome lengths. Expects names for two output files in the out_files variable. The first file is assumed to be the output file
	for the plus strand. """
	typechoices = ["reads","substitutions","deletions","startpositions","endpositions"]
	assert type in typechoices, "The option %s is not supported\n" % type
	out = defaultdict()
	if not out_files: 
		out["+"] = NGSFileWriter()
		out["-"] = NGSFileWriter()
	else:
		out["+"] = NGSFileWriter(out_files[0])
		out["-"] = NGSFileWriter(out_files[1])
		
	### writing track line to output files ###
	if name or description:
		if not name: name = "User_Supplied_Track"
		if not description: description = "User_Supplied_Track" 
		for file in out:
			if file == "+": strand = "plus"
			else: strand = "minus"
			new_name = "%s_%s_strand" % (name,strand)
			new_description = "%s_%s_strand" % (description,strand)
			out[file].writeTrackLine("bedGraph",new_name,new_description,colorbystrand) # write trackline information for visualization in the UCSC genome browser
	###
	data = ParseCountersOutput(datafile)
	data.duplicatesremoved = True
	if score:
		data.duplicatesremoved = False
	current_chromosome = str()
	chromdict = defaultdict(list)
	
	### processing the chromosome data
	chromlengthfile = open(chromosomedata,"r").readlines()
	chromdata = processChromFile(chromlengthfile)
	###
	
	normvalue = int()
	
	while True:
		lines = data.readLineByLine()
		if current_chromosome and data.chromosome != current_chromosome or not lines:
			if permillion and not normvalue:
				if not data.mapped_reads:
					sys.stderr.write("No information on the number of mapped reads could be found in the gtf file. Cannot normalize the data\n")
					normvalue = 1.0
				else:
					normvalue = float(data.mapped_reads)/1000000.0 # to normalize the data to per million reads
			else:
				normvalue = 1
			for strand in chromdict:
				start = 0
				for interval in steparray2Intervals(chromdict[strand]):
					if interval[2] != 0:
						start,end,hits = interval
						hits = float(hits)/normvalue
						out[strand].writeBedgraph(current_chromosome,start,end,hits)
			if not lines:
				break
			current_chromosome = str()
			chromdict = defaultdict(list)
			
		if not current_chromosome:
			chromosome_length = chromdata[data.chromosome]
			chromdict["+"] = np.zeros(chromosome_length)
			chromdict["-"] = np.zeros(chromosome_length)
			current_chromosome = data.chromosome
			log.write("Processing chromosome %s...\n" % current_chromosome)

		if type == "reads": 
			chromdict[data.strand][data.read_start:data.read_end] += data.number_of_reads
		elif type == "startpositions":
			if data.strand == "+":
				chromdict[data.strand][data.read_start] += data.number_of_reads
			else:
				chromdict[data.strand][data.read_end-1] += data.number_of_reads
		elif type == "endpositions":
			if data.strand == "-":
				chromdict[data.strand][data.read_start] += data.number_of_reads
			else:
				chromdict[data.strand][data.read_end-1] += data.number_of_reads
		elif type == "substitutions":
			if data.substitutions: chromdict[data.strand][data.substitutions] += data.number_of_reads
		elif type == "deletions":
			if data.deletions: chromdict[data.strand][data.deletions] += data.number_of_reads
		else:
			break

def main(): 
	parser = OptionParser(usage="usage: %prog [options] --gtf=gtf_file.gtf", version="%s" % __version__)
	files = OptionGroup(parser, "File input options")
	files.add_option("--gtf",dest="input_file",metavar="readdata.gtf",help="type the path to the gtf file data file. Be default it expects data from the standard input",default=None)
	files.add_option("-o",dest="output_file",metavar="converted",help="provide a name for an output file. A file extension or strand information is not necessary.",default=None)
	files.add_option("-c", "--chromfile",dest="chromfile",help="Location of the chromosome info file. This file should have two columns: first column is the names of the chromosomes, second column is length of the chromosomes. Default is yeast",metavar="yeast.txt",default=None)
	common = OptionGroup(parser, "bedgraph output options")
	common.add_option("-t","--type",dest="type",choices=["reads","substitutions","deletions","dropoffrates","startpositions","endpositions"],help="this tool can generate bedgraph files for reads, start positions, end positions, substitutions, deletions, reverse transcription drop-off rates (useful for iCLIP data). Please use 'substitutions','deletions','dropoffrates','startpositions','endpositions' to indicate the type of data. Default='reads'",default="reads")
	common.add_option("--count",dest="count",action="store_true",help="Takes the numbers in the 'score' column of the GTF file as the total number of reads for each position. Default is 1 for each interval.",default=False)
	common.add_option("--permillion",dest="permillion",action="store_true",help="To normalize the coverage to reads/cDNAs per million",default=False)
	common.add_option("-v", "--verbose",action="store_true",help="to print status messages to a log file",default=False)		
	ucsc = OptionGroup(parser, "These options can be used to add a track line for the UCSC genome browser")
	ucsc.add_option("-n","--name",dest="name",help="For the USCS track line: provide a track name. Default is none",default=None)
	ucsc.add_option("-d","--description",dest="description",help="For the USCS track line: provide a track description. Default is none",default=None)
	ucsc.add_option("-s","--colorstrands",dest="strands",help="select the colors for each strand. Default = 'red,blue'",default="red,blue")
	parser.add_option_group(files)
	parser.add_option_group(common)
	parser.add_option_group(ucsc)
	(options, args) = parser.parse_args()

	status = sys.stdout
	if options.verbose:
		status = open("pyGTF2bedGraph_log.txt","w")
	out = getfilename(options.input_file)
	if options.output_file:
		out = options.output_file
	file_1 = "%s_plus_strand_%s.bedgraph" % (out,options.type)
	file_2 = "%s_minus_strand_%s.bedgraph" % (out,options.type)
	outputfiles = [file_1,file_2]

	gtf2bedgraph(options.input_file,options.chromfile,out_files=outputfiles,type=options.type,permillion=options.permillion,log=status,score=options.count,name=options.name,description=options.description,colorbystrand=options.strands)
	
if __name__ == "__main__":
	main()


#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2018"
__version__		= "1.5.8"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@staffmail.ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	ParseAlignments.py
#
#
#	Copyright (c) Sander Granneman 2018
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import re
import sys
import six
import numpy as np
from collections import defaultdict
from pyCRAC.Classes.PairedReads import *
from pyCRAC.Methods import makeBinaryIndexFile,readFileIndex

#----------------------------------------------------------------- PARSECOUNTERSOUTPUT CLASS ------------------------------------------------------------------#

class ParseCountersOutput():
	"""This class parses the gtf output file generated by pyReadCounters and pyMotif"""
	def __init__(self,input_file,index=False):
		"""Initialises the class attributes"""
		if isinstance(input_file,str):					# This allows you to pipe data from the standard input. the input file would be sys.stdin
			if index:									# you can index the GTF file and do a seek.
				makeBinaryIndexFile(input_file)
				self.__fileindex = readFileIndex(input_file)
			self.input_file = open(input_file,"r")
		elif hasattr(input_file,'__iter__'):   
			self.input_file = input_file				# if it is a file object, standard input or a generator
		else:
			raise TypeError("\nCould not determine the file type\n")
		self.count			   = defaultdict(lambda: defaultdict(int))
		self.readsareclustered = False
		self.duplicatesremoved = False
		self.file_name		   = input_file
		self.read_start		   = int()
		self.read_end		   = int()
		self.genes			   = list()
		self.substitutions	   = tuple()
		self.deletions		   = tuple()
		self.chromosome		   = str()
		self.number_of_reads   = 1.0
		self.strand			   = str()
		self.attributes		   = str()
		self.mapped_reads	   = int()
		self.total_reads	   = int()
		self.single_reads	   = int()
		self.paired_reads	   = int()
		self.line			   = str()
		self.gff_type		   = str()
		self.file_header	   = list()	   # this list contains all the lines that start with a bracket. Having the headers in output files is essential.
		self.total_cDNAs	   = int()
		self.gtfdatatype	   = "reads"
		self.__clusters_regex  = re.compile("[0-9]+S\d+.\d+D\d+.\d+|[0-9]+S\d+.\d+|[0-9]+D\d+.\d+")
		self.__dels_regex	   = re.compile("([0-9]+)D")
		self.__subs_regex	   = re.compile("([0-9]+)S")
		self.__gene_regex	   = re.compile("\"(.*)\"")

	def __checkHeaderInfo(self,line):
		"""looks at header info and checks whether the data are clustered or had blocks removed"""
		if re.search("##gff-version 2",line):
			self.gff_type = "GTF"
		if re.search("clusters",line):
			self.gtfdatatype = "clusters"
			self.readsareclustered = True
			self.duplicatesremoved = False
		elif re.search("blocks",line):
			self.gtfdatatype = "cDNAs"
			self.readsareclustered = False
			self.duplicatesremoved = True
		elif re.search("peaks",line):
			self.gtfdatatype = "peaks"
			self.readsareclustered = False
			self.duplicatesremoved = True
		if re.search("total",line):
			self.file_header.append(line)
		if re.search("total number of reads",line):
			self.total_reads  = int(line.split("\t")[1])
		elif re.search("total number of mapped",line):
			self.mapped_reads = int(line.split("\t")[1])
		elif re.search("total number of single",line):
			self.single_reads = int(line.split("\t")[1])
		elif re.search("total number of paired",line):
			self.paired_reads = int(line.split("\t")[1])

	def __checkDataType(self):
		"""In case the header has been removed from the pyReadCounters or pyMotif GTF file, then this method is used to \
		read the second column (source) to see if the coordinates are for clusters, cDNAs or just reads"""
		if self.source:
			if re.search("cluster",self.source,re.I):
				self.gtfdatatype = "clusters"
				self.readsareclustered = True
				self.duplicatesremoved = False
			elif re.search("cDNA",self.source,re.I):
				self.gtfdatatype = "cDNAs"
				self.readsareclustered = False
				self.duplicatesremoved = True
			elif re.search("peak",self.source,re.I):
				self.gtfdatatype = "peaks"
				self.readsareclustered = False
				self.duplicatesremoved = True
			else:
				self.readsareclustered = False
				self.duplicatesremoved = False	

	def intervalIterator(self):
		"""Returns a generator for read or cluster positions"""
		for i in range(self.read_start, self.read_end):
			yield i

	def getGenes(self,attribute="gene_name",returnall=False):
		"""Looks for the gene_name in attributes and returns associated gene names, if any"""
		attribute_list = self.attributes.split(";")
		for i in attribute_list:
			if re.search(attribute,i):
				genes = self.__gene_regex.findall(i)
				if genes:
					return genes[0].split(",")
				else:
					return []

	def seek(self,linenumber):
		""" jumps to position in file at linenumber using an index file and index array. 
		Then it uses the next() function to start at the beginning of the line. """
		try:
			self.input_file.seek(self.__fileindex[linenumber])
			six.next(self.input_file)
		except IndexError:
			sys.stderr.write("IndexError at line:hg\t%s\nwhilst processing line:\n%s\n" % (linenumber,self.line.strip()))

	def getMutations(self,mutsfrequency=0.0):
		"""Looks for mutation information attributes and substitutions and deletions in separate tuples, if any.
		From pyReadCounters:
			if strand	== "F" : corrpos = muts - start + 1 # 1-based coordinate
			elif strand == "R" : corrpos = end	- muts	
		"""		  
		if not self.attributes:
			return [],[]
		deletions	  = list()
		substitutions = list()
		mutsfrequency = float(mutsfrequency)
		if self.gtfdatatype == "clusters":
			mutations = self.__clusters_regex.findall(self.comments)
			for muts in mutations:
				dels = list()
				subs = list()
				pos	 = re.split("[DS]",muts)
				try:
					if "D" in muts: dels = [int(pos[0]),float(pos[-1])]
					if "S" in muts: subs = [int(pos[0]),float(pos[1])]
					if dels and dels[1] >= mutsfrequency: deletions.append(dels[0]-1)			# convert them to zero based coordinates
					if subs and subs[1] >= mutsfrequency: substitutions.append(subs[0]-1) 
				except ValueError:
					sys.stderr.write("ValueError at line:\t%s. Could not process the mutations\n" % self.line.strip())
		else:
			dels = self.__dels_regex.findall(self.comments)
			subs = self.__subs_regex.findall(self.comments)
			if dels: deletions	   = [int(i)-1 for i in dels]
			if subs: substitutions = [int(i)-1 for i in subs]
		return substitutions,deletions

	def processAttributes(self,mutsfrequency,numpy=False,tuples=False,collectmutations=False):
		""" Processes the attributes in column 9 """
		self.comments = str()
		if "#" in self.attributes:
			self.comments = self.attributes.split("#")[-1]
			self.comments = self.comments.strip()
			self.comments = self.comments.strip(";")
		self.genes		= self.getGenes(attribute="gene_name")
		self.gene_ids	= self.getGenes(attribute="gene_id")
		if collectmutations:
			self.substitutions,self.deletions = self.getMutations(mutsfrequency)
			if numpy:
				self.substitutions = np.array(self.substitutions)
				self.deletions	   = np.array(self.deletions)
			elif tuples:
				self.substitutions = tuple(self.substitutions)
				self.deletions	   = tuple(self.deletions)

	def dataToDict(self,mutsfrequency=0,collectmutations=True):
		"""Opens the file and loads the information into a dictionary"""
		while self.readLineByLine(numpy=False,tuples=False,mutsseqpos=False,mutsfrequency=mutsfrequency,collectmutations=collectmutations,getattributes=True):
			interval = (self.chromosome,self.strand,self.read_start,self.read_end,tuple(self.substitutions),tuple(self.deletions))
			self.count[interval]["coverage"] = self.number_of_reads
			self.count[interval]["features"] = self.genes

	def readLineByLine(self,numpy=False,tuples=False,mutsseqpos=False,mutsfrequency=0,collectmutations=True,getattributes=True):
		"""Takes the input line, splits it and assigns values to the class attributes.
		If mutsseqpos is False then the 1-based coordinates for substitutions and deletions will be returned"""
		line_count = 0
		while True:
			self.number_of_reads = 1
			try:
				self.line = six.next(self.input_file) 
			except StopIteration:
				self.input_file.close()
				return False
			if self.line.startswith("#"):
				self.__checkHeaderInfo(self.line)
				#assert self.gff_type, "\n\nThe first line of the input file must be: ##gff-version 2\nPlease verify your input file.\n"
				continue
			#print "\t%s" % self.line.strip()
			Fld =  self.line.strip("\n").split("\t")
			self.chromosome = Fld[0]
			self.source		= Fld[1]
			self.type		= Fld[2]
			self.read_start = int(Fld[3]) - 1	# GTF annotation files contain 1-based coordinates, so 1 has to be subtracted from each value
			self.read_end	= int(Fld[4])
			self.score		= Fld[5]
			self.strand		= Fld[6]
			self.frame		= Fld[7]
			if getattributes:
				try:
					self.attributes = Fld[8]
					self.processAttributes(mutsfrequency,numpy=numpy,tuples=tuples,collectmutations=collectmutations)
				except IndexError:
					sys.stderr.write("IndexError at line %s\nAttributes are missing\n" % self.line)
			line_count += 1
			if line_count == 1: self.__checkDataType()
			if not self.readsareclustered or not self.duplicatesremoved:
				try:
					self.number_of_reads = int(self.score)
				except ValueError:
					try:
						self.number_of_reads = float(self.score)
					except ValueError:
						self.number_of_reads = 1.0
			return True

#----------------------------------------------------------------- PARSEALIGNMENTSFILE CLASS ------------------------------------------------------------------#

class ParseAlignmentFile():
	""" This class can be used to process Novoalign and SAM single and paired-end output formats
	It uses separate Novoalign and SAM modules (Novoalign.py and SAM.py) to parse each line in the alignment file"""
	def __init__(self,input_file,file_type="novo",base_quality=30,debug=None):
		self.count			   = defaultdict(lambda: defaultdict(int))
		self.file_name		   = input_file
		self.file_type		   = file_type
		self.total_reads	   = 0
		self.total_cDNAs	   = 0
		self.mapped_reads	   = 0
		self.paired_reads	   = 0
		self.single_reads	   = 0
		self.hasmutation	   = False
		self.readsareclustered = False
		self.duplicatesremoved = False
		self.readinterval	   = None
		self.line			   = str()
		self.__base_quality	   = base_quality
		self.__paired_reads	   = PairedReadCollector()
		self.__readtrackerdict = defaultdict(set)
		self.__discarded	   = None
		if debug: 
			self.__discarded = open(debug, "w")
		if re.search("novo",file_type,re.I):
			from pyCRAC.Parsers import Novoalign
			self.readinterval = Novoalign.Process_Novo(input_file,base_quality) 
		elif re.search("sam",file_type,re.I):
			from pyCRAC.Parsers import SAM
			self.readinterval = SAM.Process_Sam(input_file,base_quality)
		else:
			pass

	def printToDiscardedFile(self,line):
		""" To tell the program to print to the debug output file """
		if self.__discarded: 
			self.__discarded.write(line)

	def readisDuplicate(self,interval,muts=True):
		""" Takes a set() list and checks whether tuples with read_coordinates are already present
		in the set. If not, it will add the coordinates to the list, else it will return False.
		The reason for using sets is because membership tests are much faster than regular python lists and 
		the list doesn't need to be ordered."""
		chromosome = interval[0]
		input	   = str(interval)	# making a string out of the input coordinates makes comparison a lot easier/faster
		if not muts:
			input = str(interval[:4])
		if input in self.__readtrackerdict[chromosome]:
			return True
		else:
			self.__readtrackerdict[chromosome].add(input)
			return False

#--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
											##### COUNT ALL READS #####
#--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#

	def countReads(self,align_quality=0,align_score=0,length=1000,maximum=float("infinity"),blocks=False,force_single_end=False,unique=False,distance=1000,muts_filter=None,properpair=False,correct_seq=True):
		"""Counts the number of times reads were mapped to the same chromosomal coordinates. Before running any other counting method you have to run this one first. 
		The dictionary created by this program collects the hits for reads with the same start and end mapping position using the
		chromosome position, strand, read_start, read_end, substitutions and deletions variables (in exactly that order). For example: self.count[chromosome,strand,read_start,read_end] += 1.
		Many variables can be set: read length(i.e. length=18), how many mapped reads you want to have analyzed (maximum=10000),
		whether you only want reads with a single alignment location included (unique=True) and you can also set alignment scores and alignment qualities (align_quality=50, align_score=60).
		This method can handle both Novoalign native and BAM/SAM single and paired-end output formats"""
		self.duplicatesremoved = blocks		
		ispaired = None

		while self.readinterval.readline(correct_seq=correct_seq):
			interval = tuple()
			if properpair:
				if self.paired_reads >= maximum or not self.readinterval():
					return False
			else:
				if self.mapped_reads >= maximum or not self.readinterval():
					return False
			self.total_reads += 1

#---------------------------------------------------------------- Processing all the read filtering options ------------------------------------------------------------------#

			# checks if:
			# the read is mapped
			# the alignquality is acceptable
			# the alignscore is acceptable
			# the read has the right length
			# the read is uniquely mapped
			# the read has the desired mutations

			if not self.readinterval.ismapped()												: self.printToDiscardedFile(self.readinterval()); continue
			if align_quality and not self.readinterval.alignQualityThreshold(align_quality) : self.printToDiscardedFile(self.readinterval()); continue
			if align_score	 and not self.readinterval.alignScoreThreshold(align_score)		: self.printToDiscardedFile(self.readinterval()); continue
			if length		 and not self.readinterval.lengthThreshold(length)				: self.printToDiscardedFile(self.readinterval()); continue
			if unique		 and not self.readinterval.uniqueAlnLocation()					: self.printToDiscardedFile(self.readinterval()); continue
			if muts_filter:	 self.hasmutation = self.readinterval.filterMutations(muts_filter)		# allows filtering of mutations. If the read has the desired mutation(s) then the filterMutations method will return True, else False.

#---------------------------------------------------------------- For processing sequencing data with single end reads ------------------------------------------------#

			if force_single_end:				# in case you do not want pyReadCounters to automatically connect the paired reads.
				ispaired = False
			else:
				ispaired = self.readinterval.read_start_paired()
			if not ispaired and not properpair:
				self.line	  = self.readinterval()
				substitutions = tuple(self.readinterval.substitutions(chromosome_location=True))
				deletions	  = tuple(self.readinterval.deletions(chromosome_location=True))
				interval	  = (self.readinterval.chr_ID(),self.readinterval.strand(),self.readinterval.read_start(),self.readinterval.read_end(),substitutions,deletions)
				if blocks:
					self.count[interval]['coverage']  = 1 
				else: 
					self.count[interval]['coverage'] += 1
				self.single_reads += 1
				self.mapped_reads += 1
				continue

#--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#

										##### countReads #####

#---------------------------------------------------------------- For processing paired-end data -----------------------------------------------------------------------------#

			elif ispaired and self.readinterval.paired_sequence(distance=distance):
				self.__paired_reads.addRead(self.readinterval)
				if self.__paired_reads.numberOfReadsinList() < 2:												# to make sure that two reads have been processed before continuing.
					self.line = self.readinterval()
					continue
				if self.__paired_reads.ispaired():																# if two reads are proper pairs, then analyze them.
					self.line = "%s%s" % (self.line,self.readinterval())
					if self.__paired_reads.cDNALength() <= length:												# This bit will determine whether the read is proper length and proper align quality and/or align score
						chromosome	  = self.__paired_reads.chr_ID(read="L")
						strand		  = self.__paired_reads.strand(read="L")
						start,end	  = self.__paired_reads.coordinates(read="both")
						substitutions = tuple(set(self.__paired_reads.substitutions(read="both")))
						deletions	  = tuple(set(self.__paired_reads.deletions(read="both")))					# substitutions in the same position have to be ignored!
						interval	  = (chromosome,strand,start,end,substitutions,deletions)					# I used to have an Interval class with __slots__ but using tuples here rather than a class instance made the program 2-3 fold faster. 
						if blocks:	  # The results are stored in the self.count dictionary
							self.count[interval]['coverage']  = 1
						else:		
							self.count[interval]['coverage'] += 1
						self.paired_reads += 1
						self.mapped_reads += 2
						self.__paired_reads.reset()
					else:
						self.printToDiscardedFile(self.line)
						self.__paired_reads.reset()
				elif properpair:
					self.__paired_reads.reset()
					continue
				else:
					chromosome	  = self.__paired_reads.chr_ID(read="L")
					strand		  = self.__paired_reads.strand(read="L")
					start		  = self.__paired_reads.coordinates(read="L")[0]
					end			  = self.__paired_reads.coordinates(read="L")[1]
					substitutions = tuple(self.__paired_reads.substitutions(read="L"))
					deletions	  = tuple(self.__paired_reads.deletions(read="L"))
					interval	  = (chromosome,strand,start,end,substitutions,deletions)
					if blocks:	  # The results are stored in the self.count dictionary
						self.count[interval]['coverage']  = 1
					else:		
						self.count[interval]['coverage'] += 1
					self.single_reads += 1
					self.mapped_reads += 1
					self.__paired_reads.removeRead()
			else:
					self.printToDiscardedFile(self.readinterval())

		self.total_cDNAs = len([keys for keys in self.count])
		if self.__discarded:
			self.__discarded.close()
		return True

#--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
									##### READ_LINE_BY_LINE #####
#--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#


	def getRead(self,align_score=0,align_quality=0,length=1000,maximum=float("infinity"),blocks=False,unique=False,distance=1000,force_single_end=False,muts_filter=None,properpair=False,correct_seq=True):
		"""This method returns the start and end coordinates of a read for single end sequenced data and start and end positions of correctly paired reads. It also returns the chromosome coordinates of 
		substitutions and deletions found in the read. Insertions are removed from the read and not further considered. Many variables can be set: read length(i.e. length=18), how many mapped reads you want to have analyzed (maximum=10000),
		whether you only want reads with a single alignment locations included (unique=True) and you can also set alignment scores and alignment qualities (align_quality=50, align_score=60). The 'distance'
		indicates the maximum allowed distance between two paired reads mapped to the same chromosome. Default is 1 kb.
		This method can handle both Novoalign native and SAM output formats. Reads can be filtered for mutations. If muts_filter="delsonly", only reads with deletions are considered. Other muts_filter options
		are "subsonly", "allmuts" and "nomuts". If you set force_single_end = True then all the reads will be treated as single end reads. Useful when making multiple sequence alignments"""
		self.read_seq_ID	  = str()
		self.read_seq		  = str()
		self.mate_read_seq	  = str()
		self.read_genomic_seq = str()
		self.mate_genomic_seq = str()
		self.read_coordinates = None
		self.read_ispaired	  = False
		ispaired			  = None
		if self.total_reads == 0:
			self.duplicatesremoved	= blocks
		while self.readinterval.readline(correct_seq=correct_seq):
			interval = tuple()
			if properpair:
				if self.paired_reads >= maximum or not self.readinterval():
					return False
			else:
				if self.mapped_reads >= maximum or not self.readinterval():
					return False
			self.total_reads += 1
#--------------------------------------------------------------------- Processing all the read filtering options ---------------------------------------------------------------#

			# checks if:
			# the read is mapped
			# the alignquality is acceptable
			# the alignscore is acceptable
			# the read has the right length
			# the read is uniquely mapped
			# the read has the desired mutations

			if not self.readinterval.ismapped()												: self.printToDiscardedFile(self.readinterval()); continue
			if align_quality and not self.readinterval.alignQualityThreshold(align_quality) : self.printToDiscardedFile(self.readinterval()); continue
			if align_score	 and not self.readinterval.alignScoreThreshold(align_score)		: self.printToDiscardedFile(self.readinterval()); continue
			if length		 and not self.readinterval.lengthThreshold(length)				: self.printToDiscardedFile(self.readinterval()); continue
			if unique		 and not self.readinterval.uniqueAlnLocation()					: self.printToDiscardedFile(self.readinterval()); continue
			if muts_filter:	 self.hasmutation = self.readinterval.filterMutations(muts_filter)		# allows filtering of mutations. If the read has the desired mutation(s) then the filterMutations method will return True, else False.

#----------------------------------------------------------For processing sequencing data with single end reads ---------------------------------------------------------#

			if force_single_end:				# in case you do not want pyReadCounters to automatically connect the paired reads.
				ispaired = False
			else:
				ispaired = self.readinterval.read_start_paired()
			if not ispaired and not properpair:
				self.line		   = self.readinterval()
				self.substitutions = tuple(self.readinterval.substitutions(chromosome_location=True))
				self.deletions	   = tuple(self.readinterval.deletions(chromosome_location=True))
				self.chromosome	   = self.readinterval.chr_ID()
				self.strand		   = self.readinterval.strand()
				self.read_start	   = self.readinterval.read_start()
				self.read_end	   = self.readinterval.read_end()	
				if blocks: 
					if self.readisDuplicate((self.chromosome,self.strand,self.read_start,self.read_end,self.substitutions,self.deletions)):		 # this filter can be used to remove reads with same start and end positions on the chromosome (i.e. --blocks option)
						self.printToDiscardedFile(self.line)
						continue
				self.single_reads += 1
				self.mapped_reads += 1
				self.read_ispaired	  = False
				self.read_seq		  = self.readinterval.sequence()
				self.read_seq_ID	  = self.readinterval.seq_ID()
				self.mate_read_seq	  = None
				return True

#--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#

																			##### READ_LINE_BY_LINE#####

#---------------------------------------------------------------- For processing sequencing data with paired end reads ----------------------------------------------------------------#

			elif ispaired and self.readinterval.paired_sequence(distance=distance):	 
				self.__paired_reads.addRead(self.readinterval)
				if self.__paired_reads.numberOfReadsinList() < 2:																# to make sure that two reads have been processed before continuing.
					self.line = self.readinterval()
					continue
				if self.__paired_reads.ispaired():																				# if two reads are proper pairs, then analyze them.
					self.line = "%s%s" % (self.line,self.readinterval())
					if self.__paired_reads.cDNALength() <= length:																# This bit will determine whether the read is proper length and proper align quality and/or align score
						self.chromosome	   = self.__paired_reads.chr_ID(read="L")
						self.strand		   = self.__paired_reads.strand(read="L")
						self.read_start,self.read_end = self.__paired_reads.coordinates(read="both")
						self.substitutions = tuple(set(self.__paired_reads.substitutions(read="both")))
						self.deletions	   = tuple(set(self.__paired_reads.deletions(read="both")))
						if blocks: 
							if self.readisDuplicate((self.chromosome,self.strand,self.read_start,self.read_end,self.substitutions,self.deletions)):		# this filter can be used to remove reads with same start and end positions on the chromosome (i.e. --blocks option)
								self.printToDiscardedFile(self.line)
								self.__paired_reads.reset()																		# if block filtering is on and cDNAs with the same read coordinates already exist, empty the read list and start again
								continue
						self.paired_reads += 1
						self.mapped_reads += 2
						self.read_coordinates = interval
						self.read_ispaired	  = True
						self.read_seq_ID	  = self.__paired_reads.read_ID(read="L")
						self.read_seq		  = self.__paired_reads.sequence(read="L")
						self.mate_read_seq	  = self.__paired_reads.sequence(read="R")
						self.__paired_reads.reset()
						return True
					else:
						self.printToDiscardedFile(self.line)
						self.__paired_reads.reset()
						continue
				elif properpair:
					self.__paired_reads.reset()
					continue
				else:
					self.single_reads += 1
					self.mapped_reads += 1
					self.read_ispaired	  = False
					self.read_seq		  = self.__paired_reads.sequence(read="L")
					self.read_seq_ID	  = self.__paired_reads.read_ID(read="L")
					self.mate_read_seq	  = None
					self.chromosome		  = self.__paired_reads.chr_ID(read="L")
					self.strand			  = self.__paired_reads.strand(read="L")
					self.read_start		  = self.__paired_reads.coordinates(read="L")[0]
					self.read_end		  = self.__paired_reads.coordinates(read="L")[1]
					self.substitutions	  = tuple(self.__paired_reads.substitutions(read="L"))
					self.deletions		  = tuple(self.__paired_reads.deletions(read="L"))
					self.__paired_reads.removeRead()
					return True
#----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#

			else:
				self.printToDiscardedFile(self.readinterval())


#!/usr/bin/python

__author__		= 'Sander Granneman'
__copyright__	= 'Copyright 2019'
__version__		= '3.4'
__credits__		= ['Sander Granneman','Hywel Dunn-Davies','Grzegorz Kudla']
__maintainer__	= 'Sander Granneman'
__email__		= 'sgrannem@staffmail.ed.ac.uk'
__status__		= 'Production'

##################################################################################
#
#	Barcodes
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the 'Software'), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import time
import sys
import re
import six
import itertools
import gzip
from collections import defaultdict
from pyCRAC.Methods import sortbyvalue,getfilename
from pyCRAC.Classes.Exceptions import *


def allPossibleCombinations(barcode):
	''' returns a list of all possible combinations of a barcode with one mismatch'''
	combinations = list()
	for i in range(len(barcode)):
		newbarcode = list(barcode)
		newbarcode[i] = 'N'
		combinations.append(''.join(newbarcode))
	return combinations
	
def Npermutations(sequence):
	seq_dict		 = {'N':['A','C','G','T','N']}
	actsequence		 = [seq_dict.get(base, [base]) for base in sequence]
	all_combinations = list(map(''.join, itertools.product(*actsequence)))
	return all_combinations
	
def allPossibleNucleotideCombinations(length):
	nucleotides		= ['A','T','C','G','N']
	allcombinations = nucleotides 
	count			= 0
	while count <= length-2:
		count += 1
		allcombinations = [i+nucleotides[j] for j in range(len(nucleotides)) for i in allcombinations]
	return allcombinations
			
class BarcodeSplitter:
	''' splits fastq data based on barcode information in the 5' (!!!) adapter sequence. 
	Can also process random barcodes'''

	def __init__(self,file_type,forward,reverse,barcodefile,allowedmismatches=0,gzip=False,keepbarcode=False,search='forward'):
		self.spacer						 = '##'		# this is the character(s) that separate the header from the random barcode sequence that 
		self.randombarcodelenghts 	     = list()
		self.keepbarcode			     = keepbarcode  # in case you want to keep the barcodes in the sequencing reeds
		self.shortest_bc_length			 = int()
		self.longest_bc_length			 = int()
		self.search 					 = search
		self.total_reads				 = int()
		self.allowedmismatches			 = allowedmismatches
		self.allbarcodes				 = list()
		self.barcode_nonrandom_fragments = list()
		self.barcode_nonrandom_lengths	 = dict()
		self.barcodename				 = dict()
		self.barcode2fname				 = dict()
		self.read2barcode				 = defaultdict(str)
		self.rand_barcode_counter		 = defaultdict(int)
		self.barcode_counter			 = defaultdict(int)
		self.handles					 = dict()
		self.barcodefile				 = barcodefile
		self.forward_input				 = forward
		self.reverse_input				 = reverse
		self.fastq_f					 = None
		self.fastq_r					 = None
		self.randombarcodelengths		 = list()
		self.__filename					 = getfilename(forward)
		self.__filename_paired			 = getfilename(reverse) if reverse else None
		self.__compressoutput			 = False
		self.__allowedfiletypes			 = ['fastq','fasta','fastq.gz','fasta.gz']
		self.__outtype					 = str()
		self.filetype					 = file_type
		if self.filetype not in self.__allowedfiletypes:
			raise FormatError('\n\nCould not interpret the %s format. Please use fastq, fastq formats or gzipped fastq and fasta files	(fastq.gz or fasta.gz).\n')
		self.__processFileHandles()		# processing all the input files
		self.__processBarcodeList()		# processing the barcode names. This has to be done before creating file handles (i.e. self.__createFileHandles())
		self.__createFileHandles()		# preparing all the output files

	def __checkRandBarcodeLengths(self,firstrandombarcodes,secondrandombarcodes):
		self.randombarcodelenghts = list()
		for data in [firstrandombarcodes,secondrandombarcodes]:
			barcodelengths = set([len(i) for i in data])
			self.randombarcodelenghts.append(list(barcodelengths)[0])

	def __processFileHandles(self):
		''' Does what it says. The gzip option offers the possibility to read gzipped files '''
		if self.reverse_input:
			if self.filetype in self.__allowedfiletypes[2:]:
				self.fastq_f = gzip.open(self.forward_input,'rb')
				self.fastq_r = gzip.open(self.reverse_input,'rb')
			elif self.filetype in self.__allowedfiletypes[:2]:
				self.fastq_f = open(self.forward_input,'r')
				self.fastq_r = open(self.reverse_input,'r')
			else:
				raise FormatError('\n\nCould not interpret the %s format. Please use fastq, fastq formats or gzipped fastq and fasta files	(fastq.gz or fasta.gz).\n')
			self.handles['fastq_f'] = self.fastq_f
			self.handles['fastq_r'] = self.fastq_r
		else:
			if self.filetype in self.__allowedfiletypes[2:]:
				self.fastq_f = gzip.open(self.forward_input,'rb')
			elif self.filetype in self.__allowedfiletypes[:2]:
				self.fastq_f = open(self.forward_input,'r')
			else:
				raise FormatError('\n\nCould not interpret the %s format. Please use fastq, fastq formats or gzipped fastq and fasta files	(fastq.gz or fasta.gz).\n')
			self.handles['fastq_f'] = self.fastq_f
		
		if self.filetype == 'fasta' or self.filetype == 'fastq':
			self.__outtype = self.filetype
		else:
			self.__outtype = self.filetype[:-3]		   # removes the .gz extension
		
		if self.__compressoutput: 
			self.__outtype += '.gz'	  # add a gz extension if compressing the output files.
			
	def __createFileHandles(self):
		''' Does what it says '''
		if self.barcode_nonrandom_fragments:
			if self.__compressoutput:
				if self.reverse_input:
					self.handles['others'] = [gzip.open('%s_others.%s' % (self.__filename,self.__outtype),'wb'), gzip.open('%s_others.%s' % (self.__filename_paired,self.__outtype),'wb')]
				else:
					self.handles['others'] = gzip.open('%s_others.%s' % (self.__filename,self.__outtype),'wb')
				for names in self.allbarcodes:
					if self.reverse_input:
						self.handles[names] = [gzip.open('%s_%s_%s.%s' % (self.__filename,names,self.barcodename[names],self.__outtype), 'wb'),gzip.open('%s_%s_%s.%s' % (self.__filename_paired,names,self.barcodename[names],self.__outtype), 'wb')]
					else:
						self.handles[names] = gzip.open('%s_%s_%s.%s' % (self.__filename,names,self.barcodename[names],self.__outtype), 'wb')
			else:
				if self.reverse_input:
					self.handles['others'] = [open('%s_others.%s' % (self.__filename,self.__outtype),'w'), open('%s_others.%s' % (self.__filename_paired,self.__outtype),'w')]
				else:
					self.handles['others'] = open('%s_others.%s' % (self.__filename,self.__outtype),'w')
				for names in self.allbarcodes:
					if self.reverse_input:
						self.handles[names] = [open('%s_%s_%s.%s' % (self.__filename,names,self.barcodename[names],self.__outtype), 'w'),open('%s_%s_%s.%s' % (self.__filename_paired,names,self.barcodename[names],self.__outtype), 'w')]
					else:
						self.handles[names] = open('%s_%s_%s.%s' % (self.__filename,names,self.barcodename[names],self.__outtype), 'w')
		else:
			raise ProcessingError('\n\nYou need to process the barcode file first before creating file handles\n')
			
	def __allSeqPermutations(self,barcode):
		''' determines the number of permutations for a sequence with or without random mismatches '''
		Ncombinations	 = allPossibleCombinations(barcode)
		all_combinations = list()
		for i in Ncombinations:
			all_combinations.extend(Npermutations(i))
		return all_combinations
	
	def __nucleotideMismatches(self,fragment,trailing_nucleotides=None,mismatches=0):
		''' returns a list of all posible combinations of the barcode containing one mismatch '''
		starts		 = list()
		ends		 = list()
		combinations = list()
		if mismatches == 1:
			starts	= self.__allSeqPermutations(fragment)
		elif mismatches > 1:
			raise InputError('\n\nthis version only allows one mismatch in barcodes\n')
		else:
			starts.append(fragment)
				
		if trailing_nucleotides:
			ends.extend([''.join(i) for i in allPossibleNucleotideCombinations(trailing_nucleotides)])
		else:
			ends.append('')
		for start in starts:
			for end in ends:
				combinations.append('%s%s' % (start,end))
		return combinations
			
	def __processBarcodeList(self):
		'''processes the barcode list file and stores the information in a list'''
		randomnucs = re.compile('(N+)?([ATGC]+)+(N+)?')
		firstrandombarcodes  = list()
		secondrandombarcodes = list()
		with open(self.barcodefile, 'r') as barcode_list:
			for line in barcode_list:
				try:
					Fld = line.strip().split()		# split by anything, space or tab. In case there are spaces in the file.
					barcodeseq = Fld[0]
					samplename = Fld[1]
					self.barcodename[barcodeseq] = samplename
				except IndexError:
					continue	# in case the barcode list file has empty lines
				firstrandombarcode,index,secondrandombarcode = randomnucs.findall(barcodeseq)[0]
				firstrandombarcodes.append(firstrandombarcode)
				secondrandombarcodes.append(secondrandombarcode)
				self.barcode2fname[index] = barcodeseq
				self.allbarcodes.append(barcodeseq)
				self.barcode_nonrandom_fragments.append(index)
				self.barcode_nonrandom_lengths[barcodeseq] = len(index)
		self.__checkRandBarcodeLengths(firstrandombarcodes,secondrandombarcodes)
		self.shortest_bc_length = min([len(i) for i in self.allbarcodes])
		self.longest_bc_length	= max([len(i) for i in self.allbarcodes])
		self.barcode_nonrandom_fragments.reverse()
		for nonrandom_fragment in self.barcode_nonrandom_fragments:
			trailing_nucleotides = self.longest_bc_length - len(nonrandom_fragment) - sum(self.randombarcodelenghts)
			for sequence in self.__nucleotideMismatches(nonrandom_fragment,trailing_nucleotides,mismatches=self.allowedmismatches):
				self.read2barcode[sequence] = self.barcode2fname[nonrandom_fragment]												
	
	def processFastQFile(self):
		'''processes the fastq file'''
		assert self.allbarcodes, '\n\nYou need to process the barcode file first with processBarcodeList before you can use this method\n'
		firstrandbarcodelength  = 0
		secondrandbarcodelength = 0
		if sum(self.randombarcodelenghts) > 0:
			firstrandbarcodelength,secondrandbarcodelength = self.randombarcodelenghts
		while True:
			try:
				name	 = str(six.next(self.fastq_f).strip())
				sequence = str(six.next(self.fastq_f).strip())
				plus	 = str(six.next(self.fastq_f).strip())
				quality	 = str(six.next(self.fastq_f).strip())
				randombarcode = str()
				self.total_reads += 1
				if self.total_reads == 1 and name[0] == '>':
					raise FileTypeError('\n\nThe file is a fasta file but the selected file type is fastq. Use the flag --file_type=fasta to process fasta files\n')
				current_frag	= sequence[firstrandbarcodelength:self.longest_bc_length-secondrandbarcodelength]
				current_barcode = self.read2barcode[current_frag]
				if current_barcode in self.handles:
					if self.keepbarcode:
						length = 0
					else:
						length = firstrandbarcodelength + self.barcode_nonrandom_lengths[current_barcode] + secondrandbarcodelength 
						if firstrandbarcodelength:
							randombarcode += sequence[:firstrandbarcodelength]
						if secondrandbarcodelength:
							randombarcode += sequence[length-secondrandbarcodelength:length]
					name = '%s%s%s' % (name,self.spacer,randombarcode)
					if self.__compressoutput:
						self.handles[current_barcode].write(b'%s\n%s\n%s\n%s\n' % (name, sequence[length:], plus, quality[length:]))
					else:
						self.handles[current_barcode].write('%s\n%s\n%s\n%s\n' % (name, sequence[length:], plus, quality[length:]))
					self.barcode_counter[current_barcode] += 1
					if randombarcode:
						self.rand_barcode_counter[randombarcode] += 1
				else:
					if self.__compressoutput:
						self.handles['others'].write(b'%s\n%s\n%s\n%s\n' % (name,sequence,plus,quality))
					else:
						self.handles['others'].write('%s\n%s\n%s\n%s\n' % (name,sequence,plus,quality))
					self.barcode_counter['no_barcode'] += 1
			except StopIteration:
				for files in self.handles:
					self.handles[files].close()
				return True
	
	def processPairedEndFastQFiles(self):
		'''processes two fastq files with paired-end data. Only works with barcodes in the forward read! Forward reaction show go as first argument!!!!!'''
		assert self.allbarcodes, '\n\nYou need to process the barcode file first with processBarcodeList before you can use this method\n'
		assert self.fastq_r,'\n\nYou need to provide the name of the file containing the reverse sequencing data\n'
		firstrandbarcodelength  = 0
		secondrandbarcodelength = 0
		if sum(self.randombarcodelenghts) > 0:
			firstrandbarcodelength,secondrandbarcodelength = self.randombarcodelenghts
		while True:
			try:
				line_f = str()
				line_r = str()
				name_f, name_r	= str(six.next(self.fastq_f).strip()),str(six.next(self.fastq_r).strip())
				seq_f,	seq_r	= str(six.next(self.fastq_f).strip()),str(six.next(self.fastq_r).strip())
				plus_f, plus_r	= str(six.next(self.fastq_f).strip()),str(six.next(self.fastq_r).strip())
				qual_f, qual_r	= str(six.next(self.fastq_f).strip()),str(six.next(self.fastq_r).strip())
				randombarcode = str()
				self.total_reads += 1

				if self.search == 'reverse':
					current_frag	= seq_r[firstrandbarcodelength:self.longest_bc_length-secondrandbarcodelength]
					current_barcode = self.read2barcode[current_frag]
					if current_barcode in self.handles:
						if self.keepbarcode:
							length = 0
						else:
							length = firstrandbarcodelength + self.barcode_nonrandom_lengths[current_barcode] + secondrandbarcodelength
							if firstrandbarcodelength:
								randombarcode += seq_r[:firstrandbarcodelength]
							if secondrandbarcodelength:
								randombarcode += seq_r[length-secondrandbarcodelength:length]
						name_f = '%s%s%s' % (name_f,self.spacer,randombarcode)
						line_f = '%s\n%s\n%s\n%s\n' % (name_f,seq_f,plus_f,qual_f)
						line_r = '%s\n%s\n%s\n%s\n' % (name_r,seq_r[length:],plus_r,qual_r[length:])

						if self.__compressoutput:
							line_f = line_f.encode()
							line_r = line_r.encode()

						self.handles[current_barcode][0].write(line_f)
						self.handles[current_barcode][1].write(line_r)
						self.barcode_counter[current_barcode] += 1
						if randombarcode:
							self.rand_barcode_counter[randombarcode] += 1
					else:
						line_f = '%s\n%s\n%s\n%s\n' % (name_f, seq_f, plus_f, qual_f)
						line_r = '%s\n%s\n%s\n%s\n' % (name_r, seq_r, plus_r, qual_r)

						if self.__compressoutput:
							line_f = line_f.encode()
							line_r = line_r.encode()

						self.handles['others'][0].write(line_f)
						self.handles['others'][1].write(line_r)
						self.barcode_counter['no_barcode'] += 1

				elif self.search == 'forward':
					current_frag	= seq_f[firstrandbarcodelength:self.longest_bc_length-secondrandbarcodelength]
					current_barcode = self.read2barcode[current_frag]
					if current_barcode in self.handles:
						if self.keepbarcode:
							length = 0
						else:
							length = firstrandbarcodelength + self.barcode_nonrandom_lengths[current_barcode] + secondrandbarcodelength
							if firstrandbarcodelength:
								randombarcode += seq_f[:firstrandbarcodelength]
							if secondrandbarcodelength:
								randombarcode += seq_f[length-secondrandbarcodelength:length]
						name_f = '%s%s%s' % (name_f,self.spacer,randombarcode)
						line_f = '%s\n%s\n%s\n%s\n' % (name_f, seq_f[length:], plus_f, qual_f[length:])
						line_r = '%s\n%s\n%s\n%s\n' % (name_r, seq_r, plus_r, qual_r)
						if self.__compressoutput:
							line_f = line_f.encode()
							line_r = line_r.encode()

						self.handles[current_barcode][0].write(line_f)
						self.handles[current_barcode][1].write(line_r)
						self.barcode_counter[current_barcode] += 1

						if randombarcode:
							self.rand_barcode_counter[randombarcode] += 1
					else:
						line_f = '%s\n%s\n%s\n%s\n' % (name_f,seq_f,plus_f,qual_f)
						line_r = '%s\n%s\n%s\n%s\n' % (name_r,seq_r,plus_r,qual_r)
						if self.__compressoutput:
							line_f = line_f.encode()
							line_r = line_r.encode()

						self.handles['others'][0].write(line_f)
						self.handles['others'][1].write(line_r)
						self.barcode_counter['no_barcode'] += 1

			except StopIteration:
				for files in self.handles:
					if type(self.handles[files]) == list:
						for i in self.handles[files]:
							i.close()
					else:
						self.handles[files].close()
				return True
							
	def processFastAFile(self):
		'''processes a fasta file'''
		assert self.allbarcodes, '\n\nYou need to process the barcode file first with processBarcodeList before you can use this method\n'
		firstrandbarcodelength  = 0
		secondrandbarcodelength = 0
		if sum(self.randombarcodelenghts) > 0:
			firstrandbarcodelength,secondrandbarcodelength = self.randombarcodelenghts
		while True:
			try:
				line = str()
				name	 = str(six.next(self.fastq_f).strip())
				sequence = str(six.next(self.fastq_f).strip())
				if name[0] != '>':
					raise FormatError('\n\nthe sequence name should start with an '>'\nDid not recognize the file as fasta file\n')
				randombarcode = str()
				self.total_reads += 1
				current_frag	= sequence[firstrandbarcodelength:self.longest_bc_length-secondrandbarcodelength]
				current_barcode = self.read2barcode[current_frag]
				if current_barcode in self.handles:
					if self.keepbarcode:
						length = 0
					else:
						length = firstrandbarcodelength + self.barcode_nonrandom_lengths[current_barcode] + secondrandbarcodelength 
						if firstrandbarcodelength:
							randombarcode += sequence[:firstrandbarcodelength]
						if secondrandbarcodelength:
							randombarcode += sequence[length-secondrandbarcodelength:length]
					name = '%s%s%s' % (name,self.spacer,randombarcode)
					line = '%s\n%s\n' % (name, sequence[length:])
					if self.__compressoutput:
						line = line.encode()

					self.handles[current_barcode].write(line)
					self.barcode_counter[current_barcode] += 1

					if randombarcode:
						self.rand_barcode_counter[randombarcode] += 1
				else:
					line = '%s\n%s\n' % (name,sequence)
					if self.__compressoutput:
						line = line.encode()

					self.handles['others'].write(line)
					self.barcode_counter['no_barcode'] += 1

			except StopIteration:
				for files in self.handles:
					self.handles[files].close()
				return True

	def processPairedEndFastAFiles(self):
		'''processes two fastq files with paired-end data. Only works with barcodes in the forward read! Forward reaction show go as first argument!!!!!'''
		assert self.allbarcodes, '\n\nYou need to process the barcode file first with processBarcodeList before you can use this method\n'
		assert self.fastq_r,'\n\nYou need to provide the name of the file containing the reverse sequencing data\n'
		firstrandbarcodelength  = 0
		secondrandbarcodelength = 0
		if sum(self.randombarcodelenghts) > 0:
			firstrandbarcodelength,secondrandbarcodelength = self.randombarcodelenghts
		while True:
			try:
				line_f = str()
				line_r = str()
				name_f, name_r	= str(six.next(self.fastq_f).strip()),str(six.next(self.fastq_r).strip())
				seq_f,	seq_r	= str(six.next(self.fastq_f).strip()),str(six.next(self.fastq_r).strip())
				if name_f[0] != '>' or name_r[0] != '>':
					raise FormatError('\n\nthe sequence name should start with an '>'\nIs the file really a fasta file?\n')
				randombarcode = str()
				self.total_reads += 1
				if self.search == 'reverse':
					current_frag	= seq_r[firstrandbarcodelength:self.longest_bc_length-secondrandbarcodelength]
					current_barcode = self.read2barcode[current_frag]
					if current_barcode in self.handles:
						if self.keepbarcode:
							length = 0
						else:
							length = firstrandbarcodelength + self.barcode_nonrandom_lengths[current_barcode] + secondrandbarcodelength
							if firstrandbarcodelength:
								randombarcode += seq_r[:firstrandbarcodelength]
							if secondrandbarcodelength:
								randombarcode += seq_r[length-secondrandbarcodelength:length]
						name_f = '%s%s%s' % (name_f,self.spacer,randombarcode)
						line_f = '%s\n%s\n' % (name_f,seq_f)
						line_r = '%s\n%s\n' % (name_r,seq_r[length:])

						if self.__compressoutput:
							line_f = line_f.encode()
							line_r = line_r.encode()

						self.handles[current_barcode][0].write(line_f)
						self.handles[current_barcode][1].write(line_r)
						self.barcode_counter[current_barcode] += 1

						if randombarcode:
							self.rand_barcode_counter[randombarcode] += 1
					else:
						line_f = '%s\n%s\n' % (name_f,seq_f)
						line_r = '%s\n%s\n' % (name_r,seq_r)

						if self.__compressoutput:
							line_f = line_f.encode()
							line_r = line_r.encode()

						self.handles['others'][0].write(line_f)
						self.handles['others'][1].write(line_r)
						self.barcode_counter['no_barcode'] += 1

				elif self.search == 'forward':
					current_frag	= seq_f[firstrandbarcodelength:self.longest_bc_length-secondrandbarcodelength]
					current_barcode = self.read2barcode[current_frag]
					if current_barcode in self.handles:
						if self.keepbarcode:
							length = 0
						else:
							length = firstrandbarcodelength + self.barcode_nonrandom_lengths[current_barcode] + secondrandbarcodelength
							if firstrandbarcodelength:
								randombarcode += seq_f[:firstrandbarcodelength]
							if secondrandbarcodelength:
								randombarcode += seq_f[length-secondrandbarcodelength:length]
						name_f = '%s%s%s' % (name_f,self.spacer,randombarcode)
						line_f = '%s\n%s\n' % (name_f,seq_f[length:])
						line_r = '%s\n%s\n' % (name_r,seq_r)

						if self.__compressoutput:
							line_f = line_f.encode()
							line_r = line_r.encode()

						self.handles[current_barcode][0].write(line_f)
						self.handles[current_barcode][1].write(line_r)
						self.barcode_counter[current_barcode] += 1

						if randombarcode:
							self.rand_barcode_counter[randombarcode] += 1
					else:
						line_f = '%s\n%s\n' % (name_f,seq_f)
						line_r = '%s\n%s\n' % (name_r,seq_r)

						if self.__compressoutput:
							line_f = line_f.encode()
							line_r = line_r.encode()

						self.handles['others'][0].write(line_f)
						self.handles['others'][1].write(line_r)
						self.barcode_counter['no_barcode'] += 1

			except StopIteration:
				for files in self.handles:
					if type(self.handles[files]) == list:
						for i in self.handles[files]:
							i.close()
					else:
						self.handles[files].close()
				return True
			
	def printRandBarcodeStats(self):
		''' generates a file with random barcode statistics. '''
		if self.rand_barcode_counter:
			forwardfilename = getfilename(self.forward_input)
			random = open('%s_random_nucleotide_statistics.txt' % forwardfilename, 'w')
			random.write('# generated by Barcodes version %s, %s\n' % (__version__,time.ctime()))
			random.write('# %s\n' %(' '.join(sys.argv)) )				  
			for randseqs,value in sortbyvalue(self.rand_barcode_counter):
				random.write('%s\t%d\n' %(randseqs,value))
			random.close()
					
	def printBarcodeStats(self):
		''' generates a file with barcode statistics. '''
		if self.barcode_counter:
			forwardfilename = getfilename(self.forward_input)
			stats = open('%s_barcode_statistics.txt' % forwardfilename, 'w')
			stats.write('# generated by Barcodes version %s, %s\n' % (__version__,time.ctime()))
			stats.write('# %s\n' %(' '.join(sys.argv)) )	
			for seqs,value in sortbyvalue(self.barcode_counter):
				percentage = (float(value)/float(self.total_reads)) * 100
				stats.write('%s\t%0.2f\n' %(seqs,percentage))

							
class IndexingBarcodeSplitter(BarcodeSplitter):
	''' splits fastq data based on barcode information in the 3' (!!!) adapter sequence. 
	These are the 6-8 nucleotide illumina indexes.'''
	def __init__(self,file_type,forward,reverse,barcodefile,allowedmismatches=0,gzip=False):
		BarcodeSplitter.__init__(self,file_type,forward,reverse,barcodefile,allowedmismatches,gzip)
			
	def processFastqFileIndexes(self):
		'''sorts read data based on the index provided in the header. Example:
		@FC4290FAAXX:4:1:3:84#CAGATC/1
		The tool splits the line using the # and backslash'''
		assert self.allbarcodes, '\n\nYou need to process the barcode file first with processBarcodeList before you can use this method\n'
		index = str()
		while True:
			try:
				line = str()
				name	 = str(six.next(self.fastq_f).strip())
				sequence = str(six.next(self.fastq_f).strip())
				plus	 = str(six.next(self.fastq_f).strip())
				quality	 = str(six.next(self.fastq_f).strip())
				self.total_reads += 1
				if self.total_reads == 1 and name[0] == '>':
					raise FileTypeError('\n\nThe file is a fasta file but the selected file type is fastq. Use the flag --file_type=fasta to process fasta files\n')
				if self.randombarcode_length:
					randombarcode	= sequence[:self.randombarcode_length]
					name			= '%s%s%s' % (name,self.spacer,randombarcode)
					self.rand_barcode_counter[randombarcode] += 1
				try:
					index = re.findall('[NATGC]{6,}',name)[0]
				except IndexError:
					raise LookupError('Could not find an index for header %s\n' % name)
				current_barcode = self.read2barcode[index]

				line = '%s\n%s\n%s\n%s\n' % (name,sequence,plus,quality)

				if current_barcode in self.handles:
					length = self.randombarcode_length + self.barcode_nonrandom_lengths[current_barcode]
					self.handles[current_barcode].write(line)
					self.barcode_counter[current_barcode] += 1
				else:
					self.handles['others'].write(line)
					self.barcode_counter['no_barcode'] += 1

			except StopIteration:
				for files in self.handles:
					self.handles[files].close()
				return True
	
	def processPairedFastqFileIndexes(self):
		'''sorts paired end read data based on the index provided in the headers'''
		assert self.allbarcodes, '\n\nYou need to process the barcode file first with processBarcodeList before you can use this method\n'
		assert self.fastq_r,'\n\nYou need to provide the name of the file containing the reverse sequencing data\n'
		index = str()
		while True:
			try:
				line_f = str()
				line_r = str()
				name_f, name_r	= str(six.next(self.fastq_f).strip()),str(six.next(self.fastq_r).strip())
				seq_f,	seq_r	= str(six.next(self.fastq_f).strip()),str(six.next(self.fastq_r).strip())
				plus_f, plus_r	= str(six.next(self.fastq_f).strip()),str(six.next(self.fastq_r).strip())	  
				qual_f, qual_r	= str(six.next(self.fastq_f).strip()),str(six.next(self.fastq_r).strip())
				self.total_reads += 1
				try:
					index = re.findall('[NATGC]{6,}',name_f)[0]
				except IndexError:
					raise LookupError('Could not find an index for header %s\n' % name_f)
				current_barcode = self.read2barcode[index]

				line_f = '%s\n%s\n%s\n%s\n' % (name_f,seq_f,plus_f,qual_f)
				line_r = '%s\n%s\n%s\n%s\n' % (name_r,seq_r,plus_r,qual_r)

				if current_barcode in self.handles:
					self.handles[current_barcode][0].write(line_f)
					self.handles[current_barcode][1].write(line_r)
					self.barcode_counter[current_barcode] += 1
				else:
					self.handles['others'][0].write(line_f)
					self.handles['others'][1].write(line_r)
					self.barcode_counter['no_barcode'] += 1

			except StopIteration:
				for files in self.handles:
					if type(self.handles[files]) == list:
						for i in self.handles[files]:
							i.close()
					else:
						self.handles[files].close()
				return True
							
	def processFastAFileIndexes(self):
		'''processes a fasta file'''
		assert self.allbarcodes, '\n\nYou need to process the barcode file first with processBarcodeList before you can use this method\n'
		while True:
			try:
				line = str()
				name	 = str(six.next(self.fastq_f).strip())
				sequence = str(six.next(self.fastq_f).strip())
				if name[0] != '>':
					raise FormatError('\n\nthe sequence name should start with an '>'\nDid not recognize the file as fasta file\n')
				self.total_reads += 1
				splitheader = re.split('[#/]',name)
				if len(splitheader) == 3:			# input = @FC4290FAAXX:4:1:3:84#CAGATC/1 Should yield CAGATC
					index = splitheader[-2]		
				elif len(splitheader) == 2:			# input = @FC4290FAAXX:4:1:3:84#CAGATC	 Should yield CAGATC
					index = splitheader[-1]
				else:
					raise LookupError('Could not find an index for header %s\n' % name)								   
				current_barcode = self.read2barcode[index]

				line = '%s\n%s\n' % (name,sequence)

				if current_barcode in self.handles:
					self.handles[current_barcode].write(line)
					self.barcode_counter[current_barcode] += 1
				else:
					self.handles['others'].write(line)
					self.barcode_counter['no_barcode'] += 1

			except StopIteration:
				for files in self.handles:
					self.handles[files].close()
				return True
							
	def processPairedEndFastAFileIndexes(self):
		'''processes two fastq files with paired-end data. Only works with barcodes in the forward read! Forward reaction should go as first argument!!!!!'''
		assert self.allbarcodes, '\n\nYou need to process the barcode file first with processBarcodeList before you can use this method\n'
		while True:
			try:
				line_f = str()
				line_r = str()
				name_f, name_r	= str(six.next(self.fastq_f).strip()),str(six.next(self.fastq_r).strip())
				seq_f,	seq_r	= str(six.next(self.fastq_f).strip()),str(six.next(self.fastq_r).strip())
				if name_f[0] != '>' or name_r[0] != '>':
					raise FormatError('\n\nthe sequence name should start with an '>'\nDid not recognize the file as fasta file\n')
				self.total_reads += 1
				splitheader = re.split('[#/]',name)
				if len(splitheader) == 3:			# input = @FC4290FAAXX:4:1:3:84#CAGATC/1 Should yield CAGATC
					index = splitheader[-2]		
				elif len(splitheader) == 2:			# input = @FC4290FAAXX:4:1:3:84#CAGATC	 Should yield CAGATC
					index = splitheader[-1]
				else:
					raise LookupError('Could not find an index for header %s\n' % name)								   
				current_barcode = self.read2barcode[index]

				line_f = '%s\n%s\n' % (name_f,seq_f)
				line_r = '%s\n%s\n' % (name_r,seq_r)

				if current_barcode in self.handles:
					self.handles[current_barcode][0].write(line_f)
					self.handles[current_barcode][1].write(line_r)
					self.barcode_counter[current_barcode] += 1
				else:
					self.handles['others'][0].write(line_f)
					self.handles['others'][1].write(line_r)
					self.barcode_counter['no_barcode'] += 1

			except StopIteration:
				for files in self.handles:
					if type(self.handles[files]) == list:
						for i in self.handles[files]:
							i.close()
					else:
						self.handles[files].close()
				return True


#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "0.0.3"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@staffmail.ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyCRAC NGSFormatReaders.py
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import six

def splitter(line):
	""" This function splits the line at double quotes and returns the second object in the array"""
	line = line.replace('\"',"")
	return line.split()[1]
	

class NGSFileReader():
	""" Converts transposed numpy arrays in to various NGS output formats """
	def __init__(self,infile):
		self.file_types = ["gtf","gff","bed","bedgraph","sgr"]
		self.file_type	= str()
		if not hasattr(infile,'read'):		    # if the infile is simply a file name
			self.file = open(infile,"r")
		elif hasattr(infile,'__iter__'):   
			self.file = infile			    # if it is a file object, standard input or a generator
		else:
			raise TypeError("\nCould not determine the file type\n")
		self.chromosome = str()
		self.strand		= str()
		self.start		= int()
		self.end		= int()
		self.source		= "exon"
		self.name		= str()
		self.feature	= "interval"
		self.attributes = str()
		self.comments	= str()
		self.coverage	= int()
		self.frame		= int()
		self.comments	= list()
		self.gene_name	= str()
		self.gene_id	= str()
		self.line		= str()
		self.transcript_name = str()
		self.transcript_id	 = str()
	
	def fileComments(self):
		return "".join(self.comments)

	def resetVariables(self):
		""" resets all the class variables. Important before reading a new line """
		self.chromosome = str()
		self.strand		= str()
		self.start		= int()
		self.end		= int()
		self.source		= "exon"
		self.name		= str()
		self.feature	= "interval"
		self.attributes = str()
		self.comments	= str()
		self.coverage	= int()
		self.frame		= int()
		self.comments	= list()
		self.gene_name	= str()
		self.gene_id	= str()
		self.line		= str()
		self.transcript_name = str()
		self.transcript_id	 = str()
		
	def __retrieveAndSplitLine(self):
		""" splits any line that is tab-delimited into columns """
		self.resetVariables()
		while True:
			try:
				self.line = six.next(self.file)
			except StopIteration:
				self.file.close()
				return False
			if self.line.startswith("track"):
				self.comments.append(self.line)
				continue
			if self.line.startswith("#"):
				self.comments.append(self.line)
				continue
			else:
				return self.line.strip().split("\t")
					
	def readBedgraphLine(self):
		""" processes a line from a bedgraph file. Start positions are converted to 0-based coordinates! """
		Fld = self.__retrieveAndSplitLine()
		try:
			if Fld:
				self.chromosome = Fld[0]
				self.start		= int(Fld[1])-1
				self.end		= int(Fld[2])
				self.coverage	= Fld[4]
				return True
			else:
				return False
		except IndexError:
			sys.stderr.write("IndexError. Are you sure you selected the correct file format?\n")
			return False
					
	def readBedLine(self):
		""" processes a line from a bed file (first six columns only)"""
		Fld = self.__retrieveAndSplitLine()
		try:
			if Fld:
				self.chromosome = Fld[0]
				self.start		= int(Fld[1])
				self.end		= int(Fld[2])
				self.name		= Fld[3]
				self.score		= Fld[4]
				self.strand		= Fld[5]
				return True
			else:
				return False
		except IndexError:
			sys.stderr.write("IndexError. Are you sure you selected the correct file format?\n")
			return False
			
	def readSgrLine(self):
		""" processes a line from an sgr file. Start positions are converted to 0-based coordinates! """
		Fld = self.__retrieveAndSplitLine()
		try:
			if Fld:
				self.chromosome = Fld[0]
				self.start		= int(Fld[1])-1
				try:
					self.coverage = int(Fld[2])
				except ValueError:
					self.coverage = float(Fld[2])
				return True
			else:
				return False
		except IndexError:
			sys.stderr.write("IndexError. Are you sure you selected the correct file format?\n")
			return False

	def readGTFLine(self):	
		""" processes a line from a GTF file. Start positions are converted to 0-based coordinates! """
		Fld = self.__retrieveAndSplitLine()
		try:
			if Fld:
				self.chromosome = Fld[0]
				self.source		= Fld[1]
				self.feature	= Fld[2]
				self.start		= int(Fld[3])-1
				self.end		= int(Fld[4])
				self.score		= Fld[5]
				self.strand		= Fld[6]
				self.frame		= Fld[7]
				if len(Fld) > 8:
					self.attributes = Fld[8].strip()
					self.__processAttributes(format="gtf")
				return True
			else:
				return False
		except IndexError:
			sys.stderr.write("IndexError. Are you sure you selected the correct file format?\n")
			return False
					
	def readGFFLine(self):
		""" processes a line from a GFF3 file. Start positions are converted to 0-based coordinates! """
		Fld = self.__retrieveAndSplitLine()
		try:
			if Fld:
				self.chromosome = Fld[0]
				self.source		= Fld[1]
				self.feature	= Fld[2]
				self.start		= int(Fld[3])-1
				self.end		= int(Fld[4])
				self.score		= Fld[5]
				self.strand		= Fld[6]
				self.frame		= Fld[7]
				if len(Fld) > 8:
					self.attributes = Fld[8]
					self.__processAttributes(format="gff")
					self.comments	= self.attributes.split("#")[-1]
				return True
			else:
				return False
		except IndexError:
			sys.stderr.write("Warning! IndexError. Are you sure you selected the correct file format?\n")
	
	def __processAttributes(self,format="gtf"):
		"""takes the attributes from column 9 in the GTF or GFF file and extracts gene_name,transcript_name, etc information"""
		info			= self.attributes.split(";")
		gene_id			= str()
		transcript_id	= str()
		gene_name		= str()
		transcript_name = str()
		if format == "gtf":
			for i in info:
				i = i.strip()
				if i.startswith('gene_id'):						# gtf file format
					self.gene_id = splitter(i)		
				elif i.startswith('transcript_id'):
					self.transcript_id = splitter(i)
				elif i.startswith('gene_name'):
					self.gene_name = splitter(i)
				elif i.startswith('transcript_name'):
					self.transcript_name = splitter(i)
				
		elif format == "gff":
			for i in info:
				if i.startswith('Name='):						# gff file format
					self.gene_name = i.split("=")[1]
					self.transcript_name = self.gene_name
				elif i.startswith('ID='):
					self.gene_id = i.split("=")[1]
					self.transcript_id = self.gene_id



#!/usr/bin/python
# coding: utf-8

__author__      = "Sander Granneman"
__copyright__   = "Copyright 2019"
__version__     = "0.0.3"
__credits__     = ["Sander Granneman"]
__maintainer__  = "Sander Granneman"
__email__       = "sgrannem@ed.ac.uk"
__status__      = "Production"

##################################################################################
#
#   pySelectMotifsFromGTF.py
#
#
#   Copyright (c) Sander Granneman 2019
#   
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#   
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#   
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#
##################################################################################

import sys
import re
from optparse import *

def motif2Regex(motif):
    regex=str()
    degnuc = {
                "R" : "[A,G]",
                "Y" : "[C,T]",
                "S" : "[G,C]",
                "W" : "[A,T]",
                "K" : "[G,T]",
                "M" : "[A,C]",
                "B" : "[C,G,T]",
                "D" : "[A,G,T]",
                "H" : "[A,C,T]",
                "V" : "[A,C,G]",
                "N" : "[A,C,G,T]",
                "A" : "A",
                "T" : "T",
                "U" : "U",
                "G" : "G",
                "C" : "C"
                }
                    
    for nucs in motif:
        regex += str(degnuc[nucs.upper()])
    return regex

def selectMotifs(input_file=None,output_file=None,motif=None,score=0,length=None,mutations=False):
    count_total_lines = 0
    count_found_motifs = 0
    motif = motif2Regex(motif)
    if output_file:
        output = open(output_file, "w")
    else:
        output = sys.stdout
    while True:
        line = input_file.readline()
        if not line:
            break
        if line[0] == "#":
            output.write(line)
        else:
            count_total_lines += 1
            Fld = line.strip().split("\t")
            if length and len(Fld[2]) != length:
                continue
            if re.search(motif,Fld[2],re.I) and float(Fld[5]) >=float(score):
                output.write(line)
                count_found_motifs += 1
                                        
def main():
    parser = OptionParser(usage="usage: %prog [options] --gtf=mymotifdata.gtf -m CTTG -z 5.0 -l 4 -o CTTG.gtf", version="%s" % __version__)
    parser.add_option("--gtf",dest="gtf_file",metavar="Yourfavoritegtf.gtf",help="type the path to the gtf file that you want to use. By default it expects data from the standard input",default=None)
    parser.add_option("-o","--output",dest="output_file",metavar="FILE",help="Optional.Specify the name of the output file. Default is standard output. Make sure it has the .gtf extension!",default=None)
    parser.add_option("-m","--motif",dest="motif",metavar="KBCTTG",help="Specify the motif you want extract from the GTF file",default=None)
    parser.add_option("-z","--Z_score",dest="zscore",metavar="15.0",type="float",help="Set a minimum k-mer Z-score. Default=0",default=0)
    parser.add_option("-l","--length",dest="length",metavar="4",type="int",help="Set a limit for the k-mer length. By default it picks all the k-mer seuences that match the specified sequence/pattern",default=None)
    (options, args) = parser.parse_args()
    data = sys.stdin
    if options.gtf_file:
        data = open(options.gtf_file,"r")          
    if len(sys.argv) < 1:
        parser.error("usage: %prog [options] -f filename. Use -h or --help for options")
    if not options.motif:
        parser.error("\nYou forgot to input your motif sequence\n")
    selectMotifs(input_file=data,output_file=options.output_file,motif=options.motif,score=options.zscore,length=options.length)

if __name__ == "__main__":
    main()


#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "0.6.9"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@staffmail.ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	Coverage
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import time
import numpy as np
from pyCRAC.Classes.Exceptions import *
from pyCRAC.Methods import getfilename,numpy_overlap,reverse_strand
from pyCRAC.Parsers.ParseAlignments import ParseCountersOutput
from pyCRAC.Classes.NGSFormatWriters import NGSFileWriter
from collections import defaultdict

class BinCollector():
	def __init__(self):
		""" File input should be the gtf file generated by PyCounters with gene_name attributes in it!"""
		np.seterr(all='ignore')					# ignore zero division errors
		self.dataarray	  = defaultdict(list)
		self.numberofbins = int()
		self.sequence	  = str()
		self.annotation	  = str()
		self.filename	  = str()
		self.range		  = int()
		self.cumulativedata = None
		self.__feature_a  = ["coding","genomic","5UTR","3UTR","TSS","5end","3end","CDSstart","CDSend"]
		self.__feature_b  = ["intron","CDS","exon","5ss","3ss"]
		self.__types	  = ["intervals","deletions","substitutions"]
		self.__orientations = ["sense","anti_sense"]
		self.orientation  = "sense"
		if self.sequence == "CDSstart" or self.sequence == "CDSend":
			assert self.annotation == "protein_coding", "You cannot used CDSstart and CDSend without setting the -a flag to protein_coding"
		
	def findBinOverlap(self,datafile,gtf,printbins=[],sequence="genomic",orientation="sense",minfeatlength=100000000,maxfeatlength=0,ignorestrand=False,out_file=None,ranges=0,overlap=1):
		""" returns those lines that overlap with specific bin numbers. Has a bedtools-like function. 
		Allows you to find those intervals that overlap with, for example, 5' or 3' ends of genomic features."""
		assert orientation in self.__orientations, "The orientation %s is not supported, please choose from the following:\t%s\n" % (orientation,", ".join(self.__orientations))
		assert sequence in self.__feature_a or sequence in self.__feature_b, "Can not determine what sequence coordinates you want to use.\nOptions are:\t%s\tand\t%s\n.\nPlease check your input\n" % (", ".join(self.__feature_a),", ".join(self.__feature_b))
		assert self.numberofbins, "No bin number was provided\n"
		self.sequence	  = sequence	
		self.filename	  = datafile
		self.range		  = ranges
		data = ParseCountersOutput(datafile)
		outfile = NGSFileWriter()
		if out_file:
			outfile = NGSFileWriter(out_file)
		outfile.write("##gff-version 2\n# generated by BinCollector, %s\n# %s\n" % (time.ctime(),' '.join(sys.argv)))

		list_of_tuples = defaultdict(list)

		#### reading the GTF file line by line ####
		
		while data.readLineByLine(numpy=True,collectmutations=False):							# reads line by line over the GTF data file or stdin and returns the coordinates as a numpy array
			if ignorestrand:
				strand = None
			elif orientation == "sense":
				strand = data.strand
			else:
				strand = reverse_strand(data.strand)
			search_results = []
			if (data.chromosome,strand) not in list_of_tuples:		# list_of_tuples is as it says...the tuples contain gene names and chromosomal positions.
				try:
					list_of_tuples[(data.chromosome,strand)] = gtf.chromosomeGeneCoordIterator(data.chromosome,annotation=self.annotation,numpy=True,strand=strand,ranges=ranges,sequence=sequence)
				except AssertionError:
					sys.stderr.write("Can't get chromosomeGeneCoordinates for chromosome %s and sequence %s" % (data.chromosome,sequence))

			if len(list_of_tuples[(data.chromosome,strand)]) > 0:	# if there are any genes on the chromosome...
				search_results = numpy_overlap(list_of_tuples[(data.chromosome,strand)],data.read_start,data.read_end-1,overlap=overlap)
			else:
				continue
			if search_results:
				for gene in search_results:
					if sequence in self.__feature_a:
						itercoordinates = [gtf.geneIterCoordinates(gene,coordinates=sequence,ranges=ranges)]
					elif sequence in self.__feature_b:
						itercoordinates = gtf.geneIterCoordinates(gene,coordinates=sequence,ranges=ranges)
					else:
						raise LookupError("Could not generate iteration coordinates, please check your input\n")
					for coordinates in itercoordinates:
						if type(coordinates) is np.ndarray and coordinates.any():
							feature_length = len(coordinates)
							if feature_length >= minfeatlength and feature_length <= maxfeatlength and feature_length >= self.numberofbins:
								if gtf.strand(gene) == "-": 
									coordinates = coordinates[::-1]		# reverse the itercoordinates if the feature is on the minus strand
								hits = np.array_split((coordinates < data.read_end) & (coordinates >= data.read_start),self.numberofbins)
								if True in [True in i for i in hits[printbins[0]:printbins[-1]]]:	   # if there are any hits in the selected bins: print the line!
									gene_name = ",".join(search_results)
									gene_id = ",".join([gtf.gene2orf(gene) for gene in search_results])
									outfile.writeGTF(data.chromosome,data.source,data.type,data.read_start,data.read_end,data.score,data.strand,gene_name=gene_name,gene_id=gene_id,comments=data.comments)

	def findOverlap(self,datafile,gtf,sequence="genomic",orientation="sense",minfeatlength=0,maxfeatlength=100000000,ignorestrand=False,out_file=None,ranges=0,overlap=1):
		""" returns those lines that overlap with specific features such as exons, coding sequences, introns, UTRs, etc. Has a bedtools-like function. """
		assert orientation in self.__orientations, "Error! Orientation %s is not recognized. Please choose %s" % (orientation, " or ".join(self.__orientations))
		assert sequence in self.__feature_a or sequence in self.__feature_b, "Can not determine what sequence coordinates you want to use.\nOptions are:\t%s\tand\t%s\n.\nPlease check your input\n" % (", ".join(self.__feature_a),", ".join(self.__feature_b))
		self.sequence = sequence
		self.filename = datafile
		self.range	  = ranges
		data = ParseCountersOutput(datafile)
		outfile = NGSFileWriter()
		if out_file:
			outfile = NGSFileWriter(out_file)
		outfile.write("##gff-version 2\n# generated by BinCollector, %s\n# %s\n" % (time.ctime(),' '.join(sys.argv)))
		
		list_of_tuples = defaultdict(list)

		#### reading the GTF file line by line ####
		while data.readLineByLine(numpy=True,collectmutations=False):	# reads line by line over the GTF data file or stdin and returns the coordinates as a numpy array
			if ignorestrand:
				strand = None
			elif orientation == "sense":
				strand = data.strand
			else:
				strand = reverse_strand(data.strand)
			search_results = []
			if (data.chromosome,strand) not in list_of_tuples:		    # list_of_tuples is as it says...the tuples contain gene names and chromosomal positions.
				try:
					list_of_tuples[(data.chromosome,strand)] = gtf.chromosomeGeneCoordIterator(data.chromosome,annotation=self.annotation,numpy=True,strand=strand,ranges=ranges,sequence=sequence)
				except AssertionError:
					sys.stderr.write("Can't get chromosomeGeneCoordinates for chromosome %s and sequence %s" % (data.chromosome,sequence))
					
			if len(list_of_tuples[(data.chromosome,strand)]) > 0:	    # if there are any genes on the chromosome...
				search_results = numpy_overlap(list_of_tuples[(data.chromosome,strand)],data.read_start,data.read_end-1,overlap=overlap)
			else:
				continue
			if search_results:
				gene_name = ",".join(search_results)
				gene_id = ",".join([gtf.gene2orf(gene) for gene in search_results])
				outfile.writeGTF(data.chromosome,data.source,data.type,data.read_start,data.read_end,data.score,data.strand,gene_name=gene_name,gene_id=gene_id,comments=data.comments)
	
	def countFeatureDensities(self,datafile,gtf,sequence="genomic",feature="intervals",minfeatlength=0,maxfeatlength=100000000,orientation="sense",ignorestrand=False,out_file=None,ranges=0,overlap=1,unique=False):
		""" counts the gene hits without dividing the gene lengths into bins"""
		assert feature in self.__types, "The feature %s is not supported, please choose from the following:\t%s\n" % (feature,", ".join(self.__types))
		assert orientation in self.__orientations, "The orientation %s is not supported, please choose from the following:\t%s\n" % (orientation,", ".join(self.__orientations))
		assert sequence in self.__feature_a or sequence in self.__feature_b, "Can not determine what sequence coordinates you want to use.\nOptions are:\t%s\tand\t%s\n.\nPlease check your input\n" % (", ".join(self.__feature_a),", ".join(self.__feature_b))
		self.sequence	 = sequence
		self.filename	 = datafile
		self.range		 = ranges
		self.type		 = feature
		self.orientation = orientation
		collectmuts = True
		if feature == "intervals":
			collectmuts = False

		data = ParseCountersOutput(datafile)
		list_of_tuples = defaultdict(list)

		#### reading the GTF file line by line. This program assumes the intervals are sorted by chromosome! ####

		while data.readLineByLine(numpy=True,collectmutations=collectmuts):					# reads line by line over the GTF data file or stdin and returns the coordinates as a numpy array
			if ignorestrand:
				strand = None
			elif orientation == "sense":
				strand = data.strand
			else:
				strand = reverse_strand(data.strand)
			search_results = []
			if (data.chromosome,strand) not in list_of_tuples:		# list_of_tuples is as it says...the tuples contain gene names and chromosomal positions.
				try:
					list_of_tuples[(data.chromosome,strand)] = gtf.chromosomeGeneCoordIterator(data.chromosome,annotation=self.annotation,numpy=True,strand=strand,ranges=ranges,sequence=sequence)
				except AssertionError:
					sys.stderr.write("Can't get chromosomeGeneCoordinates for chromosome %s and sequence %s" % (data.chromosome,sequence))
			if len(list_of_tuples[(data.chromosome,strand)]) > 0:	# if there are any genes on the chromosome...
				search_results = set(numpy_overlap(list_of_tuples[(data.chromosome,strand)],data.read_start,data.read_end-1,overlap=overlap))
			else:
				continue
			if search_results:
				for gene in search_results:
					if sequence in self.__feature_a:
						itercoordinates = [gtf.geneIterCoordinates(gene,coordinates=sequence,ranges=ranges)]
					elif sequence in self.__feature_b:
						itercoordinates = gtf.geneIterCoordinates(gene,coordinates=sequence,ranges=ranges)
					else:
						raise LookupError("Could not generate iteration coordinates, please check your input\n")
					for nr,coordinates in enumerate(itercoordinates):
						if type(coordinates) is np.ndarray and coordinates.any():
							feature_length = len(coordinates)
							if feature_length >= minfeatlength and feature_length <= maxfeatlength and feature_length >= self.numberofbins:
								if (gene,nr) not in self.dataarray: 
									self.dataarray[(gene,nr)] = np.zeros(len(coordinates))
								if gtf.strand(gene) == "-": 
									coordinates = coordinates[::-1]			# reverse the itercoordinates if the feature is on the minus strand
								if unique:
									data.number_of_reads = 1
								if feature == "intervals":
									self.dataarray[(gene,nr)][(coordinates < data.read_end) & (coordinates >= data.read_start)] += data.number_of_reads
								elif feature == "substitutions":
									if data.substitutions.any(): 
										self.dataarray[(gene,nr)][np.in1d(coordinates,data.substitutions)] += data.number_of_reads
								elif feature == "deletions":
									if data.deletions.any():
										self.dataarray[(gene,nr)][np.in1d(coordinates,data.deletions)] += data.number_of_reads
		
	def binFeatureDensities(self,numberofbins=20):
		""" divides the features in an equal number of bins """
		assert self.dataarray, "No data to be binned!\n"
		if not self.numberofbins:
			self.numberofbins = numberofbins
		for (gene,nr) in sorted(self.dataarray):
			if len(self.dataarray[(gene,nr)]) >= self.numberofbins:
				self.dataarray[(gene,nr)] = np.array([i.sum() for i in np.array_split(self.dataarray[(gene,nr)],self.numberofbins)])
			else:
				del(self.dataarray[(gene,nr)])
				
	def __createFileHandles(self,file_type,file_extension="txt"):
		""" does what it says """
		filename   = str()
		annotation = "all"
		if self.annotation:
			annotation = self.annotation
		filename = "%s_%s_%s_%s_%s" % (getfilename(self.filename),file_type,annotation,self.sequence,file_extension)
		return filename

	def printBinCountingResults(self,out_file=None,cumulative=False,normalize=False,permillion=False):
		""" prints the binCounter results in a pileup or table format. This method can be called repeatedly to append data to an existing file if the same output file name is used """
		assert self.dataarray, "\n\nNo data to be printed. Did you run the binCounter method?\nDid you set the maximum sequence length too high?\nPlease double check your settings\n"
		normvalue = 1.0
		if permillion:
		    assert self.data.mapped_reads, "Could not calculate hits permillion as I could not tell how many reads were mapped to the genome\n"
		if cumulative:
			assert self.numberofbins, "The data has not been divided into bins yet\n"
			file_type = "cumulative_densities"
			if out_file:
				filename = out_file
			else:
				filename = "%s_%s.pileup" % (getfilename(self.filename),file_type)
			outputfile = open(filename,"w")
			cumulativedata = np.zeros(self.numberofbins)
			sumofall = np.sum(list(self.dataarray.values()))
			for (gene,nr) in self.dataarray:
				if type(self.dataarray[(gene,nr)]) is np.ndarray and self.dataarray[(gene,nr)].any(): 
					if normalize:
						data = self.dataarray[(gene,nr)]/sumofall
						data[np.isnan(data)] = 0.0
					elif permillion:
					    data = self.dataarray[(gene,nr)]/(float(self.data.mapped_reads)/1000000.0)
					else: 
						data = self.dataarray[(gene,nr)]
					cumulativedata += data
			outputfile.write("%s\n" % "\t".join(["%.6f" % i for i in cumulativedata]))
			outputfile.close()
		else:
			filename = str()
			if out_file:
				filename = out_file
			else:
				filename = self.__createFileHandles("%s_%s" % (self.orientation,self.type),file_extension="txt")
			outputfile = open(filename,"w")
			for (gene,nr) in self.dataarray:
				if type(self.dataarray[(gene,nr)]) is np.ndarray and self.dataarray[(gene,nr)].any(): 
					if normalize: 
						data = self.dataarray[(gene,nr)]/sum(self.dataarray[(gene,nr)])
					elif permillion:
					    data = self.dataarray[(gene,nr)]/float(self.data.mapped_reads)
					else:
						data = self.dataarray[(gene,nr)]
					data[np.isnan(data)] = 0.0
					if data.any():
						string = "\t".join(["%.8f" % i for i in data])
						outputfile.write("%s\t%s\n" % (gene,string))
		outputfile.close()



#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2018"
__version__		= "0.0.3"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyExtractLinesFromGTF.py
#
#
#	Copyright (c) Sander Granneman 2018
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import re
from optparse import *

def getGenes(line,attribute="gene_name"):
	"""Looks for the gene_name in attributes and returns associated gene names, if any"""
	gene_regex	   = re.compile("\"(.*)\"")
	attributefield = list()
	genes		   = list()
	try:
		attributefield = line.split("\t")[8]
	except IndexError:
		return []
	attribute_list = attributefield.split(";")
	for i in attribute_list:
		if re.search("\\b%s\\b" % attribute,i):	# only attributes with the whole word "gene_name" or "gene_id" will be selected!
			genes = gene_regex.findall(i)
			if genes:
				return genes[0].split(",")
	if not genes:
		for i in attribute_list:
			if re.search("gene_id",i):
				genes = gene_regex.findall(i)
				if genes:
					return genes[0].split(",")
	else:
		return []

def grep(input_file,output_file,genes_list,remove=False):
	gene_line_dict = dict()
	genes_list = set(genes_list)
	genes = set()
	output_file.write("##gff-version 2\n# %s\n" % " ".join(sys.argv))
	for line in input_file:
		if line[0] == "#":
			continue
		genes = set(getGenes(line))
		if remove:
			if genes_list & genes:
				continue
			else:
				output_file.write(line)
		else:
			if genes_list & genes:
				output_file.write(line)

def extractGenesFromGTF(genes_list,inputfile,outputfile,remove=False):
	genes = list()
	with open(genes_list, "r") as genes_list:
		for line in genes_list:
			genes.append(line.strip())
	grep(inputfile,outputfile,genes,remove=remove)

def main():
	parser = OptionParser(usage="usage: %prog --gtf gtf_file -g genes.list", version="%prog "+__version__)
	parser.add_option("--gtf",dest="gtf_file",metavar="Yourfavoritegtf.gtf",help="type the path to the gtf file that you want to use. By default it expects data from the standard input.",default=None)
	parser.add_option("-g", "--genes_file", dest="genes",help="name of your gene list or annotations list file (1 column)", metavar="FILE",default=None)
	parser.add_option("-o","--outfile",dest="outfile",help="type the name and path of the file you want to write the output to. Default is standard output",default=None)
	parser.add_option("-a","--attribute", dest="attribute",choices=["gene_name","gene_id","transcript_name","transcript_id"], help="from which attribute do you want to extract names? Choices: gene_name, gene_id, transcript_name, transcript_id", default="gene_name")
	parser.add_option("-v", dest="remove",action="store_true",help="similar to grep -v option. Remove the genes from the GTF that are in the gene list",default=False)
	(options, args) = parser.parse_args()
	data = sys.stdin
	out	 = sys.stdout
	if options.outfile:
		out = open(options.outfile,"w")
	if options.gtf_file:
		data = open(options.gtf_file,"r")
	extractGenesFromGTF(options.genes,data,out,remove=options.remove)

if __name__ == "__main__":
	main()



#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright "
__version__		= "0.0.5"
__credits__		= ["Sander Granneman"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyFastqDuplicateRemover.py
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import re
import six
from collections import defaultdict
from optparse import *

def getfilename(filepath):
	""" Returns the file name without path and extension"""
	return filepath.split("/")[-1].split(".")[0]

def trackDuplicates(file_input):
	""" keeps track of the cDNA duplicates using both random barcode sequences (if available)
	and the nucleotide sequence. Counts the number of duplicates. Returns a dictionary """
	randbarcode = None
	header		= str()
	tracker		= defaultdict(int)
	splitter	= "##"	# the character(s) that divide the header and random barcode sequence
	while True:
		try:
			name_f = six.next(file_input).strip()
			seq_f  = six.next(file_input).strip()
			if name_f.startswith("@"):	# for fastq files
				six.next(file_input)	# To make sure that the next round it starts at a header again in the next round
				six.next(file_input)	# The program doesn't care about the quality values				
				if re.search(splitter,name_f):	# If there is a random barcode sequence in the header of the forward sequence...
					randbarcode = re.split(splitter,name_f)[-1].split("@")[0]  # split again with @ in case these are joined header from paired end fastq data
			elif name_f.startswith(">"): # for fasta files
				if re.search(splitter,name_f):	# If there is a random barcode sequence in the header of the forward sequence...
					randbarcode = re.split(splitter,name_f)[-1].split(">")[0]  # split again with > in case these are joined header from paired end fasta data
			
			tracker[randbarcode,seq_f] += 1
					
		except StopIteration:
			break
	
	return tracker

def trackPairedEndDuplicates(forward,reverse):
	""" Tracks duplicates in Paired-End data. Returns a dictionary """
	randbarcode = None
	header		= str()
	tracker		= defaultdict(int)
	splitter	= "##"	# the character(s) that divide the header and random barcode sequence
	while True:
		try:
			name_f, name_r	= six.next(forward).strip(),six.next(reverse).strip()
			seq_f,	seq_r	= six.next(forward).strip(),six.next(reverse).strip()
			if name_f.startswith("@"):
				six.next(forward)		# To make sure that the next round it starts at a header again in the next round
				six.next(reverse)		# The program doesn't care about the quality values
				six.next(forward)
				six.next(reverse)
			if re.search(splitter,name_f):	# If there is a random barcode sequence in the header of the forward sequence...
				randbarcode = re.split(splitter,name_f)[-1]
			else:
				randbarcode = "##"
			tracker[randbarcode,seq_f,seq_r] += 1
			
		except StopIteration:
			break
	
	return tracker
	
def printFastaFile(tracker,output=None):
	""" prints the results in fasta format. Header includes random barcode sequence and number of duplicates """
	if output:
		file_output = open(output,"w")
	else:
		file_output = sys.stdout		# if it doesn't get
		
	for rank,key in enumerate(tracker,start=1):
		randbarcode,sequence = key
		if randbarcode:
			file_output.write(">%s_%s_%s\n%s\n" % (rank,randbarcode,tracker[randbarcode,sequence],sequence))
		else:
			file_output.write(">%s_%s\n%s\n" % (rank,tracker[randbarcode,sequence],sequence))

def printPairedEndFastaFile(tracker,output=None):
	""" prints the results in fasta format for Paired-End data. Header includes random barcode sequence and number of duplicates """
	file_name = str()
	if output:
		file_name = getfilename(output)
	else:
		file_name = "collapsed"
	output_f = open("%s_1.fasta" % (file_name), "w")
	output_r = open("%s_2.fasta" % (file_name), "w")
	for rank,key in enumerate(tracker,start=1):
		randbarcode,sequence_f,sequence_r = key
		if randbarcode:
			output_f.write(">%s_%s_%s/1\n%s\n" % (rank,randbarcode,tracker[randbarcode,sequence_f,sequence_r],sequence_f))
			output_r.write(">%s_%s_%s/2\n%s\n" % (rank,randbarcode,tracker[randbarcode,sequence_f,sequence_r],sequence_r))
		else:
			output_f.write(">%s_%s/1\n%s\n" % (rank,tracker[randbarcode,sequence_f,sequence_r],sequence_f))
			output_r.write(">%s_%s/2\n%s\n" % (rank,tracker[randbarcode,sequence_f,sequence_r],sequence_r))

def main():
	parser = OptionParser(usage="usage: %prog [options] -f forwardfile -r reversefile -o collapsed.fasta", version="%s" % __version__)
	files = OptionGroup(parser,"File input options")
	parser.add_option("-f", "--input_file", dest="infile_forward",help="name of the FASTQ or FASTA input file. Default is standard input", metavar="FILE",default=None)
	parser.add_option("-r", "--reverse_input_file", dest="infile_reverse",help="name of the paired (or reverse) FASTQ or FASTA input file", metavar="FILE",default=None)
	parser.add_option("-o", "--output_file", dest="outfile",help="Provide the path and name of the fasta output file. Default is standard output. For paired-end data just provide a file name without file extension (!)",metavar="FILE",default=None)
	(options, args) = parser.parse_args()
	infile_f  = sys.stdin
	outfile	  = None
	if options.outfile: 
		outfile = options.outfile
	if options.infile_forward and not options.infile_reverse: 
		infile_f = open(options.infile_forward,"r")
		datadict = trackDuplicates(infile_f)
		printFastaFile(datadict,outfile)
	elif options.infile_forward and options.infile_reverse: 
		infile_f = open(options.infile_forward,"r")
		infile_r = open(options.infile_reverse,"r")
		datadict = trackPairedEndDuplicates(infile_f,infile_r)
		printPairedEndFastaFile(datadict,outfile)
	else:
		datadict = trackDuplicates(infile_f)
		printFastaFile(datadict,outfile)

if __name__ == "__main__":
	main()



#!/usr/bin/python

__author__		= "Sander Granneman"
__copyright__	= "Copyright 2019"
__version__		= "0.1.0"
__credits__		= ["Sander Granneman","Grzegorz Kudla"]
__maintainer__	= "Sander Granneman"
__email__		= "sgrannem@ed.ac.uk"
__status__		= "Production"

##################################################################################
#
#	pyMotif.py
#
#
#	Copyright (c) Sander Granneman 2019
#	
#	Permission is hereby granted, free of charge, to any person obtaining a copy
#	of this software and associated documentation files (the "Software"), to deal
#	in the Software without restriction, including without limitation the rights
#	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#	copies of the Software, and to permit persons to whom the Software is
#	furnished to do so, subject to the following conditions:
#	
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#	
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#	THE SOFTWARE.
#
##################################################################################

import sys
import re
import six
from optparse import *
from pyCRAC.Classes.Motifs import *
from pyCRAC.Parsers import GTF2

def main():
	"""This program can only be used with an alignment file that has been generated using the genome as reference sequence!"""
	parser = OptionParser(usage="usage: %prog [options] -f filename --gtf=yeast.gtf --tab=yeast.tab", version="%s" % __version__)
	files = OptionGroup(parser, "File input options")
	files.add_option("-f", "--input_file", dest="filename",help="Provide the path to an interval gtf file. By default it expects data from the standard input.", metavar="intervals.gtf",default=None)
	files.add_option("-o", "--output_file", dest="output_file",help="Use this flag to override the standard file names. Do NOT add an extension.", default=None)
	files.add_option("--file_type",dest="file_type",choices=['bed','gff','gtf'], help=SUPPRESS_HELP,default='gtf')
	files.add_option("--gtf",dest="annotation_file",metavar="annotation_file.gtf",help="type the path to the gtf annotation file that you want to use",default=None)
	files.add_option("--tab",dest="tab_file",metavar="tab_file.tab",help="type the path to the tab file that contains the genomic reference sequence",default=None) 
	specific = OptionGroup(parser, "pyMotif specific options")
	specific.add_option("--k_min", dest="k_length_min",type="int",metavar="4",help="this option allows you to set the shortest k-mer length. Default = 4.",default=4)
	specific.add_option("--k_max", dest="k_length_max",type="int",metavar="6",help="this option allows you to set the longest k-mer length. Default = 8.",default=8)
	specific.add_option("-n","--numberofkmers",dest="motif_number",type="int",metavar=100,help="choose the maximum number of enriched k-mer sequences you want to have reported in output files. Default = 1000",default=100000)
	common = OptionGroup(parser, "pyCRAC common options")	
	common.add_option("-a","--annotation",dest="annotation",help="select which annotation (i.e. protein_coding, ncRNA, sRNA, rRNA,snoRNA,snRNA, depending on the source of your GTF file) you would like to focus your search on. Default = all annotations",metavar="protein_coding",default=None) 
	common.add_option("-r","--range",dest="range",type="int",metavar="100",help="allows you to add regions flanking the genomic feature. If you set '-r 50' or '--range=50', then the program will add 50 nucleotides to each feature on each side regardless of whether the GTF file has genes with annotated UTRs.",default=False)
	common.add_option("-v","--verbose",dest="verbose",action="store_true",help="prints all the status messages to a file rather than the standard output",default=False)
	common.add_option("--overlap",dest="overlap",help="sets the number of nucleotides a motif has to overlap with a genomic feature before it is considered a hit. Default =  1 nucleotide",type="int",metavar="1",default=1)
	common.add_option("--zip", dest="zip",metavar="FILE", help="use this option to compress all the output files in a single zip file", default=None)
	parser.add_option_group(files)
	parser.add_option_group(specific)
	parser.add_option_group(common)
	# the --out option is used by Galaxy to override default output file names. Four file names need to be included. First file will contain Z-scores, second the random motifs, third the motifs found in the data and fourth file will contain motifs found in genes.'
	# recommended file names: 'Z_scores.txt','random_k-mer_hit_list.txt','data_k-mer_hit_list.txt','top_k-mers_in_features.gtf'
	(options, args) = parser.parse_args()
	status = sys.stdout
	if options.verbose:
		status = open("pyMotif_log.txt","w")
	if not options.filename:
		options.filename = sys.stdin
	if len(sys.argv) < 1:
		parser.error("no input files detected, use -h or --help for instructions on how to use the program\n")
	if not options.annotation_file:
		parser.error("you forgot to input the GTF annotation file\n")
	if not options.tab_file:
		parser.error("you forgot to input the .tab file\n")
	
	### Making GTF2 parser object and parsing GTF annotation file ###
	status.write("processing gtf annotation and genomic sequence files...\n")
	gtf = GTF2.Parse_GTF()
	gtf.read_GTF(options.annotation_file,ranges=options.range,source=options.annotation,transcripts=False)
	gtf.read_TAB(options.tab_file)
	###

	### creating a FindMotifs object
	motifs = FindMotifs()
	### setting lengths of the k-mers to be extracted and set the annotation
	motifs.setKmerLength(k_min=options.k_length_min, k_max=options.k_length_max)
	motifs.annotation = options.annotation
	
	### processing the read data and finding motifs:
	status.write("finding motifs...\n")
	motifs.findMotifs(gtf,options.filename,min_overlap=options.overlap)
	###
	
	status.write("calculating k-mer Z-scores...\n")
	motifs.calcMotifZscores()

	motifs.printMotifZscores(output_file=options.output_file,max_motifs=options.motif_number)
	status.write("printing k-mers hits...\n")
	motifs.printRandKmers(output_file=options.output_file,max_motifs=options.motif_number)
	status.write("printing randomly extracted k-mer hits...\n")
	motifs.printExperimentalKmers(output_file=options.output_file,max_motifs=options.motif_number)
	status.write("printing k-mer gtf file...\n")
	motifs.printMotifGTFfile(output_file=options.output_file,max_motifs=options.motif_number)
	status.write("DONE!...\n")	

	if options.zip:					
		import zipfile
		status.write("compressing files...\n")
		file = zipfile.ZipFile(options.zip,mode="w",compression=zipfile.ZIP_DEFLATED)
		for files in motifs.file_list:
			file.write(files)
		file.close()	
		for files in motifs.file_list:
			try:
				os.remove(files)
			except OSError:
				pass   

if __name__ == "__main__":
	main()